from __future__ import annotations

import logging
import os
import warnings
from pathlib import Path
from typing import Literal
from typing import Optional
from typing import Sequence
from typing import Tuple

import numpy
import tqdm
from silx.io import open
from silx.io import utils
from silx.io.dictdump import dicttoh5
from silx.io.dictdump import h5todict
from silx.io.url import DataUrl
from sklearn import decomposition
from sklearn.exceptions import ConvergenceWarning

from darfix import __version__
from darfix.core.dimension import AcquisitionDims
from darfix.core.dimension import find_dimensions_from_metadata
from darfix.core.imageRegistration import apply_opencv_shift
from darfix.core.imageRegistration import shift_detection
from darfix.core.imageStack import ImageStack
from darfix.core.mapping import calculate_RSM_histogram
from darfix.core.mapping import compute_magnification
from darfix.core.mapping import compute_moments
from darfix.core.mapping import compute_rsm
from darfix.core.positioners import Positioners
from darfix.core.rocking_curves import fit_1d_data
from darfix.core.rocking_curves import fit_2d_data
from darfix.core.rocking_curves import fit_3d_data
from darfix.core.roi import apply_2D_ROI
from darfix.core.roi import apply_3D_ROI
from darfix.core.state_of_operation import Operation
from darfix.core.state_of_operation import StateOfOperations
from darfix.core.utils import NoDimensionsError
from darfix.core.utils import TooManyDimensionsForRockingCurvesError
from darfix.decomposition.nica import NICA
from darfix.io import utils as io_utils_legacy
from darfix.processing.imageOperations import threshold_removal

from ..math import Vector3D
from .transformation import Transformation

_logger = logging.getLogger(__file__)

CHUNK_SIZE = 100


def _read_data_by_chunk(detector_url: DataUrl | str) -> numpy.ndarray:
    if isinstance(detector_url, DataUrl):
        detector_url = detector_url.path()
    with open(detector_url) as detector_data:
        raw_data = numpy.ndarray(detector_data.shape, detector_data.dtype)
        n_frames = detector_data.shape[0]
        if n_frames > CHUNK_SIZE:
            n_chunk = n_frames // CHUNK_SIZE
            # Iterate by chunk of CHUNK_SIZE
            for i_chunk in tqdm.tqdm(
                range(n_chunk), desc="Read data by chunk", total=n_chunk
            ):
                chunk_slice = slice(i_chunk * CHUNK_SIZE, (i_chunk + 1) * CHUNK_SIZE)
                raw_data[chunk_slice] = detector_data[chunk_slice]
            # Last frames
            chunk_slice = slice(n_chunk * CHUNK_SIZE, None)
            raw_data[chunk_slice] = detector_data[chunk_slice]
            return raw_data
        # In case there is only a few frames (less than CHUNK_SIZE)
        return utils.h5py_read_dataset(detector_data)


class ImageDataset(ImageStack):
    """Class to define a darfix dataset.

    :param _dir: Global directory to use and save all the data in the different
        operations.
    :param raw_data: If not None, set detector data to this numpy array.
    :param detector_url: data url of the detector data. Either `raw_data` or `detector_url` should be given.
    :param dims: Dimensions dictionary
    :param transformation: Axes to use when displaying the images
    :param metadata_url: url to the metadata
    :param positioners: positioners metadata instance
    .. warning::

        When loading the full dataset will be store to memory.
    """

    def __init__(
        self,
        _dir: str,
        raw_data: numpy.ndarray | None = None,
        detector_url: str | DataUrl | None = None,
        dims: AcquisitionDims | None = None,
        transformation: numpy.ndarray | None = None,
        title: str | None = None,
        metadata_url: DataUrl | str | None = None,
        positioners: Positioners | None = None,
    ):

        if raw_data is None:
            raw_data = _read_data_by_chunk(detector_url)
        super().__init__(raw_data)

        self._frames_intensity = []
        self.moments_dims = {}
        self.state_of_operations = StateOfOperations()
        self._dir = _dir
        self._transformation = transformation
        self._title = title or ""

        if dims is None:
            self.__dims = AcquisitionDims()
        else:
            assert isinstance(
                dims, AcquisitionDims
            ), "Attribute dims has to be of class AcquisitionDims"
            self.__dims = dims
        # Keys: dimensions names, values: dimensions values
        self._dimensions_values = {}

        if positioners is not None and metadata_url is not None:
            raise ValueError("Cannot set both positioners and metadata_url")

        if positioners is not None:
            self._positioners = positioners
        elif metadata_url is not None:

            self._positioners = Positioners(metadata_url)
        else:
            # No metadata in Dark dataset for instance
            self._positioners = None

    def stop_operation(self, operation):
        """
        Method used for cases where threads are created to apply functions to the dataset.
        If method is called, the flag concerning the stop is set to 0 so that if the concerned
        operation is running in another thread it knows to stop.

        :param int operation: operation to stop
        :type int: Union[int, `Operation`]
        """
        if self.state_of_operations.is_running(operation):
            self.state_of_operations.stop(operation)

    @property
    def transformation(self):
        return self._transformation

    @transformation.setter
    def transformation(self, value):
        self._transformation = value

    @property
    def title(self):
        return self._title

    @property
    def metadata_url(self) -> DataUrl | None:
        if self._positioners is None:
            return None
        return self._positioners.url

    @property
    def dir(self):
        return self._dir

    @property
    def metadata_dict(self) -> dict[str, numpy.ndarray]:
        if self._positioners is None:
            return {}
        return self._positioners.data

    def compute_frames_intensity(self, kernel=(3, 3), sigma=20):
        """
        Returns for every image a number representing its intensity. This number
        is obtained by first blurring the image and then computing its variance.
        """
        _logger.info("Computing intensity per frame")
        io_utils_legacy.advancement_display(0, self.nframes, "Computing intensity")
        frames_intensity = []
        with self.state_of_operations.run_context(Operation.PARTITION):
            for i in range(self.nframes):
                import cv2

                if not self.state_of_operations.is_running(Operation.PARTITION):
                    return
                frames_intensity += [
                    cv2.GaussianBlur(self.as_array3d(i), kernel, sigma).var()
                ]
                io_utils_legacy.advancement_display(
                    i + 1, self.nframes, "Computing intensity"
                )
            self._frames_intensity = frames_intensity
            return frames_intensity

    def partition_by_intensity(
        self,
        bins: Optional[int] = None,
        bottom_bin: Optional[int] = None,
        top_bin: Optional[int] = None,
    ):
        """
        :param bins: number of bins to used for computing the frame intensity histogram
        :param bottom_bin: index of the bins to retrieve bottom threshold filter value. If not provided, there will be no bottom threshold (default).
        :param top_bin: index of the bins to retrieve top threshold filter value. If not provided, there will be no top threshold (default).

        Function that computes the data from the set of urls.
        If the filter_data flag is activated it filters the data following the next:
        -- First, it computes the intensity for each frame, by calculating the variance after
        passing a gaussian filter.
        -- Second, computes the histogram of the intensity.
        -- Finally, saves the data of the frames with an intensity bigger than a threshold.
        The threshold is set to be the second bin of the histogram.
        """
        frames_intensity = (
            self._frames_intensity
            if self._frames_intensity
            else self.compute_frames_intensity()
        )
        if frames_intensity is None:
            return
        _, bins = numpy.histogram(
            frames_intensity, self.nframes if bins is None else bins
        )
        frames_intensity = numpy.asanyarray(frames_intensity)
        if top_bin is None:
            top_bin = len(bins) - 1
        if bottom_bin is None:
            bottom_bin = 0

        bottom_threshold = frames_intensity >= bins[bottom_bin]
        top_threshold = frames_intensity <= bins[top_bin]
        threshold = numpy.array(
            [a and b for a, b in zip(bottom_threshold, top_threshold)]
        )
        return numpy.flatnonzero(threshold), numpy.flatnonzero(~threshold)

    def copy(
        self,
        new_dir: str | None = None,
        new_data: numpy.ndarray | None = None,
        new_positioners: Positioners | None = None,
        new_transformation: numpy.ndarray | None = None,
        new_dims: AcquisitionDims | None = None,
    ) -> ImageDataset:
        new_dataset = ImageDataset(
            _dir=new_dir if new_dir is not None else self._dir,
            raw_data=new_data if new_data is not None else self._data,
            positioners=(
                new_positioners if new_positioners is not None else self._positioners
            ),
            transformation=(
                new_transformation
                if new_transformation is not None
                else self._transformation
            ),
            dims=new_dims if new_dims is not None else self.__dims,
            title=self._title,
        )
        new_dataset.reshape_data()
        return new_dataset

    @property
    def dims(self):
        return self.__dims

    @dims.setter
    def dims(self, _dims):
        if not isinstance(_dims, AcquisitionDims):
            raise TypeError(
                "Dimensions dictionary has " "to be of class `AcquisitionDims`"
            )
        self.__dims = _dims

    def zsum(self, indices=None, dimension=None) -> numpy.ndarray:
        data = self.get_filtered_data(indices, dimension)
        return data.sum(axis=0)

    def reshape_data(self):
        """
        Function that reshapes the data to fit the dimensions.
        """
        if self.dims.ndim == 0:
            # No dimension => reshape to simple 3d array
            dims_shape = (self.nframes,)
        else:
            dims_shape = self.__dims.shape

        try:
            self._data = self.as_array3d().reshape(dims_shape + self.frame_shape)
        except Exception:
            dimensions = " ".join(self.__dims.get_names())
            raise ValueError(
                f"Dimensions {dimensions} have {dims_shape} points while there are {self.nframes} images. Try using other tolerance or step values."
            )

    def find_dimensions(self, tolerance: float = 1e-9) -> None:
        """
        Call core.dimension.find_dimensions_from_metadata to set __dims

        :param tolerance: Tolerance used to find dimensions
        """
        self.__dims = find_dimensions_from_metadata(self.metadata_dict, tolerance)

    def get_metadata_values(self, key: str, indices=None) -> numpy.ndarray:
        if key not in self.metadata_dict:
            # The key does not exist -> return an array of the desired size and with NaN value
            return numpy.full(self.nframes, numpy.nan)
        if indices is None:
            return self.metadata_dict[key]
        return self.metadata_dict[key][indices]

    def get_dimensions_values(self, indices=None):
        """
        Returns all the metadata values of the dimensions.
        The values are assumed to be numbers.

        :returns: array_like
        """
        if not self._dimensions_values or indices is not None:
            for dimension in self.__dims.values():
                self._dimensions_values[dimension.name] = self.get_metadata_values(
                    key=dimension.name, indices=indices
                )
        return self._dimensions_values

    def apply_roi(
        self, origin=None, size=None, center=None, indices=None, roi_dir=None
    ):
        """
        Applies a region of interest to the data.

        :param origin: Origin of the roi
        :param size: [Height, Width] of the roi.
        :param center: Center of the roi
        :type origin: Union[2d vector, None]
        :type center: Union[2d vector, None]
        :type size: Union[2d vector, None]
        :param indices: Indices of the images to apply background subtraction.
            If None, the roi is applied to all the data.
        :type indices: Union[None, array_like]
        :param roi_dir: Directory path for the new dataset
        :type roi_dir: str
        :returns: dataset with data with roi applied.
            Note: To preserve consistence of shape between images, if `indices`
            is not None, only the data modified is returned.
        :rtype: Dataset
        """

        new_data = apply_3D_ROI(self.as_array3d(), origin, size, center)

        transformation = self.transformation
        if transformation is not None:
            transformation = Transformation(
                transformation.kind,
                apply_2D_ROI(transformation.x, origin, size, center),
                apply_2D_ROI(transformation.y, origin, size, center),
                transformation.rotate,
            )

        return self.copy(new_data=new_data, new_transformation=transformation)

    def z_sum_along_axis(self, axis_to_keep: int) -> numpy.ndarray:
        """
        Compute the zsum for each point of the `axis_to_keep`.

        For instance, let's take self.data with shape (a,b,c,m,n)` where (m,n) is the frame shape in pixels and (a,b,c) the shape of the 3 motors :

        If `axis_to_keep` = 0, it means we compute we sum pixel values in axis 0 and 2 in order to obtain an array with shape (a, m, n).
        If `axis_to_keep` = 1, shape of the resulting array would be (b, m, n)
        If `axis_to_keep` = 2, shape of the resulting array would be (c, m, n)

        Note that the method assumes self.data.ndim >= 3.
        """
        axis_array = tuple(
            axis for axis in range(self.data.ndim - 2) if axis_to_keep != axis
        )
        return self.data.sum(axis_array)

    def find_shift(
        self, selected_axis: int | None = None, steps=50, indices=None
    ) -> tuple[float, float]:
        """
        Find shift of the data or part of it.

        :param int selected_axis: specify one motor (axis). The method try to find shift along this axis.
        :param array_like indices: Boolean index list with True in the images to apply the shift to.
        :returns: tuple with shift x and y
        """
        # TODO : indices are not used. Maybe it is possible to use it for performance or maybe it is better to simply remove it.
        data = (
            self.z_sum_along_axis(selected_axis)
            if selected_axis is not None
            else self.as_array3d()
        )

        return shift_detection(data, steps)

    def apply_shift(
        self,
        shift: tuple[float, float],
        axis: int | None = None,
        shift_approach: str | None = None,
        indices=None,
    ):
        """
        Apply shift of the data or part of it and save new data into disk.

        :param array_like shift: Shift per frame.
        :param int axis: Select one axis (motor) where to apply the shift. If None, the shift id not applied to one axis but to the whole stack.
        :param Literal['fft', 'linear'] shift_approach: Method to use to apply the shift.
        :param array_like indices: Boolean index list with True in the images to apply the shift to.
            If None, the hot pixel removal is applied to all the data.
        """
        assert len(shift) > 0, "Shift list can't be empty"

        if shift_approach is None:
            shift_approach = "linear"

        if axis is None:

            # Cumulative shift: frame 0 shift is 0, frame 1 shift is `shift`, frame 2 shift is `2*shift`, ...
            cumulative_shift_per_frames = numpy.outer(
                shift, numpy.arange(self.nframes)
            ).T
        else:
            cumulative_shift_per_frames = numpy.outer(
                shift, numpy.indices(self.scan_shape)[axis]
            ).T

        data = self.as_array3d()

        for img_index, shift in enumerate(
            tqdm.tqdm(
                cumulative_shift_per_frames,
                "Apply shift",
                len(cumulative_shift_per_frames),
            )
        ):
            if numpy.isclose(shift[0], 0) and numpy.isclose(shift[1], 0):
                # No need to apply a (0,0) shift
                continue
            if indices is not None and img_index not in indices:
                # Not really optimized but the simplest way i found to take indices into account.
                continue
            data[img_index] = apply_opencv_shift(
                data[img_index],
                shift,
                shift_approach=shift_approach,
            )

    def find_and_apply_shift(
        self,
        selected_axis: int | None = None,
        steps=100,
        indices=None,
        shift_approach: str = "linear",
    ):
        """
        Find the shift of the data or part of it and apply it.

        :param int selected_axis: specify one motor (axis). The method try to find shift along this axis.
        :param float h_step: See `core.imageRegistration.shift_detection`
        :param Union['fft', 'linear'] shift_approach: Method to use to apply the shift.
        :param array_like indices: Indices of the images to find and apply the shift to.
        :param Literal['fft', 'linear'] shift_approach: Method to use to apply the shift.
        """
        shift = self.find_shift(selected_axis, steps, indices=indices)
        self.apply_shift(
            shift, selected_axis, indices=indices, shift_approach=shift_approach
        )

    def _waterfall_nmf(
        self, num_components, iterations, vstep=None, hstep=None, indices=None
    ):
        """
        This method is used as a way to improve the speed of convergence of
        the NMF method. For this, it uses a waterfall model where at every step
        the output matrices serve as input for the next.
        That is, the method starts with a smaller resized images of the data,
        and computes the NMF decomposition. The next step is the same but with
        bigger images, and using as initial H and W the precomputed matrices.
        The last step is done using the actual size of the images.
        This way, the number of iterations with big images can be diminished, and
        the method converges faster.

        :param int num_components: Number of components to find.
        :param array_like iterations: Array with number of iterations per step of the waterfall.
            The size of the array sets the size of the waterfall.
        :param Union[None, array_like] indices: If not None, apply method only to indices of data.
        """

        from skimage.transform import resize

        W = None
        H = None
        shape = self.frame_shape
        first_size = (shape / (len(iterations) + 1)).astype(int)
        size = first_size

        _logger.info("Starting waterfall NMF")

        for i in range(len(iterations)):
            if vstep:
                v_step = vstep * (len(iterations) - i)
            if hstep:
                h_step = hstep * (len(iterations) - i)
            H, W = self.nmf(
                num_components, iterations[i], W=W, H=H, vstep=v_step, hstep=h_step
            )
            size = first_size * (i + 2)
            H2 = numpy.empty((H.shape[0], size[0] * size[1]))
            for row in range(H.shape[0]):
                H2[row] = resize(H[row].reshape((i + 1) * first_size), size).flatten()
            H = H2

        H = resize(H, (num_components, shape[0] * shape[1]))

        return H, W

    def pca(self, num_components=None, indices=None, return_vals=False):
        """
        Compute Principal Component Analysis on the data.
        The method, first converts, if not already, the data into an hdf5 file object
        with the images flattened in the rows.

        :param num_components: Number of components to find.
            If None, it uses the minimum between the number of images and the
            number of pixels.
        :type num_components: Union[None, int]
        :param indices: If not None, apply method only to indices of data, defaults to None
        :type indices: Union[None, array_like], optional
        :param return_vals: If True, returns only the singular values of PCA, else returns
            the components and the mixing matrix, defaults to False
        :type return_vals: bool, optional

        :return: (H, W): The components matrix and the mixing matrix.
        """
        bss_dir = os.path.join(self.dir, "bss")
        os.makedirs(bss_dir, exist_ok=True)

        model = decomposition.PCA(n_components=num_components)

        W = model.fit_transform(self.as_array2d((indices)))

        H, vals, W = model.components_, model.singular_values_, W

        return vals if return_vals else (H, W)

    def nica(
        self,
        num_components,
        chunksize=None,
        num_iter=500,
        error_step=None,
        indices=None,
    ):
        """
        Compute Non-negative Independent Component Analysis on the data.
        The method, first converts, if not already, the data into an hdf5 file object
        with the images flattened in the rows.

        :param num_components: Number of components to find
        :type num_components: Union[None, int]
        :type chunksize: Union[None, int], optional
        :param num_iter: Number of iterations, defaults to 500
        :type num_iter: int, optional
        :param error_step: If not None, find the error every error_step and compares it
            to check for convergence. TODO: not able for huge datasets.
        :param indices: If not None, apply method only to indices of data, defaults to None
        :type indices: Union[None, array_like], optional

        :return: (H, W): The components matrix and the mixing matrix.
        """

        model = NICA(
            self.as_array2d(),
            num_components,
            chunksize,
            indices=indices,
        )
        model.fit_transform(max_iter=num_iter, error_step=error_step)
        return numpy.abs(model.H), numpy.abs(model.W)

    def nmf(
        self,
        num_components,
        num_iter=100,
        error_step=None,
        waterfall=None,
        H=None,
        W=None,
        vstep=100,
        hstep=1000,
        indices=None,
        init=None,
    ):
        """
        Compute Non-negative Matrix Factorization on the data.
        The method, first converts, if not already, the data into an hdf5 file object
        with the images flattened in the rows.

        :param num_components: Number of components to find
        :type num_components: Union[None, int]
        :param num_iter: Number of iterations, defaults to 100
        :type num_iter: int, optional
        :param error_step: If not None, find the error every error_step and compares it
            to check for convergence, defaults to None
            TODO: not able for huge datasets.
        :type error_step: Union[None, int], optional
        :param waterfall: If not None, NMF is computed using the waterfall method.
            The parameter should be an array with the number of iterations per
            sub-computation, defaults to None
        :type waterfall: Union[None, array_like], optional
        :param H: Init matrix for H of shape (n_components, n_samples), defaults to None
        :type H: Union[None, array_like], optional
        :param W: Init matrix for W of shape (n_features, n_components), defaults to None
        :type W: Union[None, array_like], optional
        :param indices: If not None, apply method only to indices of data, defaults to None
        :type indices: Union[None, array_like], optional

        :return: (H, W): The components matrix and the mixing matrix.
        """
        bss_dir = os.path.join(self.dir, "bss")
        os.makedirs(bss_dir, exist_ok=True)

        model = decomposition.NMF(
            n_components=num_components, init=init, max_iter=num_iter
        )
        X = self.as_array2d(indices)

        if numpy.any(X[:, :] < 0):
            _logger.warning("Setting negative values to 0 to compute NMF")
            X[X[:, :] < 0] = 0
        if H is not None:
            X = X.astype(H.dtype)
        elif W is not None:
            X = X.astype(W.dtype)
        with warnings.catch_warnings():
            warnings.simplefilter("always", ConvergenceWarning)
            W = model.fit_transform(X, W=W, H=H)
        return model.components_, W

    def nica_nmf(
        self,
        num_components,
        chunksize=None,
        num_iter=500,
        waterfall=None,
        error_step=None,
        vstep=100,
        hstep=1000,
        indices=None,
    ):
        """
        Applies both NICA and NMF to the data. The init H and W for NMF are the
        result of NICA.
        """
        H, W = self.nica(num_components, chunksize, num_iter, indices=indices)

        # Initial NMF factorization: X = F0 * G0
        W = numpy.abs(W)
        H = numpy.abs(H)

        return self.nmf(
            min(num_components, H.shape[0]),
            num_iter,
            error_step,
            waterfall,
            H,
            W,
            vstep,
            hstep,
            indices=indices,
            init="custom",
        )

    def apply_moments(
        self,
        indices=None,
    ):
        """
        Compute the COM, FWHM, skewness and kurtosis of the data for very dimension.

        :param indices: If not None, apply method only to indices of data, defaults to None
        :type indices: Union[None, array_like], optional
        """

        if not self.dims.ndim:
            raise NoDimensionsError("apply_moments")
        for axis, dim in self.dims.items():
            # Get motor values per image of the stack
            values = self.get_dimensions_values(indices)[dim.name]
            self.moments_dims[axis] = numpy.asarray(
                compute_moments(values, self.as_array3d(indices)),
                dtype=numpy.float64,
            )

        return self.moments_dims

    def apply_fit(
        self,
        indices=None,
        int_thresh: float | None = 15.0,
        method: str | None = None,
        _dir: str | None = None,
    ) -> Tuple[ImageDataset, numpy.ndarray]:
        """
        Fits the data around axis 0 and saves the new data into disk.

        :param indices: Indices of the images to fit.
            If None, the fit is done to all the data.
        :type indices: Union[None, array_like]
        :param int_thresh: see `mapping.fit_pixel`
        :type int_thresh: Union[None, float]
        :returns: dataset with data of same size as `self.data` but with the
            modified images. The urls of the modified images are replaced with
            the new urls.
        """
        if not self.dims.ndim:
            raise NoDimensionsError("apply_fit")

        new_dataset = self.copy(new_data=self.data.copy())

        with self.state_of_operations.run_context(Operation.FIT):
            data = new_dataset.as_array3d(indices)
            # Fit can only be done if rocking curves are at least of size 3
            if len(data) < 3:
                raise ValueError(
                    f"Fit can only be done if rocking curves are at least of dimensionality 3. Got {len(data)}"
                )

            if self.dims.ndim == 1:
                dim0 = self.dims.get(0)
                data[:], maps = fit_1d_data(
                    data=data,
                    values=self.get_metadata_values(dim0.name, indices),
                    int_thresh=int_thresh,
                    method=method,
                )
            elif self.dims.ndim == 2:
                # TODO: `indices` seem to be handled differently here ??
                dim0 = self.dims.get(0)
                dim1 = self.dims.get(1)
                values = [
                    self.get_metadata_values(key=dim0.name),
                    self.get_metadata_values(key=dim1.name),
                ]
                data[:], maps = fit_2d_data(
                    data=self.as_array3d(),
                    values=values,
                    shape=(dim0.size, dim1.size),
                    int_thresh=int_thresh,
                    indices=indices,
                    method=method,
                )
            elif self.dims.ndim == 3:
                # TODO: `indices` seem to be handled differently here ??
                dim0 = self.dims.get(0)
                dim1 = self.dims.get(1)
                dim2 = self.dims.get(2)
                data[:], maps = fit_3d_data(
                    data=self.as_array3d(),
                    motor_values=[
                        self.get_metadata_values(key=dim0.name),
                        self.get_metadata_values(key=dim1.name),
                        self.get_metadata_values(key=dim2.name),
                    ],
                    shape=(dim0.size, dim1.size, dim2.size),
                    int_thresh=int_thresh,
                    indices=indices,
                    method=method,
                )
            else:
                raise TooManyDimensionsForRockingCurvesError()

        return (
            new_dataset,
            maps,
        )

    def compute_transformation(
        self,
        d: float,
        kind: Literal["magnification", "rsm"] = "magnification",
        rotate: bool = False,
        topography_orientation: int | None = None,
        center: bool = True,
    ):
        """
        Computes transformation matrix.
        Depending on the kind of transformation, computes either RSM or magnification
        axes to be used on future widgets.

        :param d: Size of the pixel
        :param kind: Transformation to apply, either 'magnification' or 'rsm'
        :param rotate: To be used only with kind='rsm', if True the images with
            transformation are rotated 90 degrees.
        :param topography: To be used only with kind='magnification', if True
            obpitch values are divided by its sine.
        """

        if not self.dims:
            raise NoDimensionsError("compute_transformation")

        H, W = self.frame_shape
        self.rotate = rotate

        def get_dataset(name) -> float | numpy.array | None:
            if name in self._positioners.data:
                return numpy.median(self._positioners.data[name])
            if name not in self._positioners.constants:
                raise ValueError(
                    f"No dataset named '{name}' found. Unable to compute transformation"
                )
            return self._positioners.constants[name]

        mainx = get_dataset("mainx")

        if kind == "rsm":
            if self.dims.ndim != 1:
                raise ValueError(
                    "RSM transformation matrix computation is only for 1D datasets. Use kind='magnification' or project the dataset first."
                )
            ffz = get_dataset("ffz")
            x, y = compute_rsm(H, W, d, ffz, -mainx, rotate)
        else:
            obx = get_dataset("obx")
            obpitch = get_dataset("obpitch")
            x, y = compute_magnification(
                H, W, d, obx, obpitch, -mainx, topography_orientation, center
            )
        self.transformation = Transformation(kind, x, y, rotate)

    def project_data(self, dimension: Sequence[int], indices=None, _dir=None):
        """
        Applies a projection to the data.
        The new Dataset will have the same size as the chosen dimension, where
        the data is projected on.

        :param dimension: Dimensions to project the data onto
        :type dimension: array_like
        :param indices: Indices of the images to use for the projection.
            If None, the projection is done using all data.
        :type indices: Union[None, array_like]
        :param str _dir: Directory filename to save the new data
        """

        if not self.dims:
            raise NoDimensionsError("project_data")

        dims = AcquisitionDims()
        if len(dimension) == 1:
            axis = self.dims.ndim - dimension[0] - 1
            dim = self.dims.get(dimension[0])
            data = []
            for i in range(dim.size):
                _sum = self.zsum(indices=indices, dimension=[dimension[0], i])
                if len(_sum):
                    data += [_sum]
            data = numpy.array(data)
            dims.add_dim(0, dim)
        elif len(dimension) == 2:
            axis = int(numpy.setdiff1d(range(self.dims.ndim), dimension)[0])
            dim1 = self.dims.get(dimension[0])
            dim2 = self.dims.get(dimension[1])
            dims.add_dim(0, dim1)
            dims.add_dim(1, dim2)
            data = []
            for i in range(dim1.size):
                for j in range(dim2.size):
                    _sum = self.zsum(indices=indices, dimension=[dimension, [i, j]])
                    if len(_sum):
                        data += [_sum]
            data = numpy.array(data)
        else:
            raise ValueError("Only 1D and 2D projection is allowed")

        dim = self.dims.get(axis)

        return self.copy(new_data=data, new_dims=dims)

    def compute_rsm(
        self,
        Q: Vector3D,
        a: float,
        map_range: float,
        pixel_size: float,
        units: Literal["poulsen", "gorfman"] | None = None,
        n: Vector3D | None = None,
        map_shape: Vector3D | None = None,
        energy: float | None = None,
        transformation: Transformation | None = None,
    ):
        diffry = self.get_metadata_values("diffry")
        if transformation is None:
            transformation = self.transformation

        if transformation is None:
            raise ValueError(
                "Transformation has to be computed first using the `compute_transformation` method"
            )

        return calculate_RSM_histogram(
            data=self.as_array3d(),
            diffry_values=diffry,
            twotheta=transformation.y,
            eta=transformation.x,
            Q=Q,
            a=a,
            map_range=map_range,
            units=units,
            map_shape=map_shape,
            n=n,
            E=energy,
        )

    def apply_binning(self, scale, _dir=None):
        from darfix.tasks.binning import Binning  # avoid cyclic import

        _dir = self.dir if _dir is None else _dir
        task = Binning(
            inputs={
                "dataset": self,
                "output_dir": _dir,
                "scale": scale,
            }
        )
        task.run()
        return task.outputs.dataset

    def recover_weak_beam(self, n, indices=None):
        """
        Set to zero all pixels higher than n times the standard deviation across the stack dimension

        :param n: Increase or decrease the top threshold by this fixed value.
        :type n: float
        :param indices: Indices of the images to use for the filtering.
            If None, the filtering is done using all data.
        :type indices: Union[None, array_like]
        """
        std = numpy.std(self.as_array3d(indices).view(numpy.ndarray), axis=0)

        dataset = self.copy(new_data=self.data.copy())
        threshold_removal(dataset.as_array3d(indices), top=n * std)
        return dataset

    @staticmethod
    def load_from_output_dir(output_dir: str, filename: str = "data.h5"):
        file = Path(output_dir) / filename
        h5dict = h5todict(file)

        if "positioners" in h5dict:
            positioners = Positioners(
                DataUrl(file_path=file, data_path="/positioners", scheme="silx"),
                h5dict["positioners"],
            )
        else:
            positioners = None

        if "dimensions" in h5dict:
            # h5dict values are numpy dict but we want a string
            for dim in h5dict["dimensions"].values():
                dim["name"] = str(dim["name"])
            # h5dict keys are str but we want a integer
            raw_dimensions = {
                int(str_axis): dim for str_axis, dim in h5dict["dimensions"].items()
            }
            dimensions = AcquisitionDims.from_dict(raw_dimensions)
        else:
            dimensions = None

        loaded_dataset = ImageDataset(
            output_dir,
            raw_data=h5dict["raw_data"],
            dims=dimensions,
            transformation=h5dict.get("transformation", None),
            title=h5dict.get("title", None),
            positioners=positioners,
        )
        loaded_dataset.reshape_data()
        return loaded_dataset

    def save_in_output_dir(self, filename: str = "data.h5"):
        file = Path(self._dir) / filename
        if "RAW_DATA" in file.parts:
            raise PermissionError(
                f"Write in RAW_DATA dir of ESRF is not allowed: {file}"
            )
        h5dict = {
            "raw_data": self.as_array3d(),
            "dimensions": self.dims.to_dict(),
            "positioners": self.metadata_dict,
            "transformation": self.transformation,
            "title": self.title,
            "writer": {"name": "darfix", "version": __version__},
        }
        dicttoh5(h5dict, file, mode="a", update_mode="modify")



from __future__ import annotations

from silx.gui import qt

from .utils import create_completer


class FileSelector(qt.QWidget):
    """Widget to select a file"""

    sigFileChanged = qt.Signal()

    def __init__(
        self,
        parent: qt.QWidget | None = None,
        label: str = "file:",
        filters=None,
    ) -> None:
        super().__init__(parent)
        self._dialogFileMode = qt.QFileDialog.AnyFile
        self._dialogNameFilters = None  # ["HDF5 file *.h5 *.hdf5 *.nx *.nexus", "nxs"]

        self.setLayout(qt.QHBoxLayout())

        # label
        self._label = qt.QLabel(text=label)
        self.layout().addWidget(self._label)

        # file path QLE and completer
        self._filePath = qt.QLineEdit("")
        self.layout().addWidget(self._filePath)
        self.completer = create_completer(filters)
        self._filePath.setCompleter(self.completer)

        # select button
        self._select = qt.QPushButton("select")
        self.layout().addWidget(self._select)

        # connect signal / slot
        self._select.released.connect(self._selectFile)
        self._filePath.editingFinished.connect(self.sigFileChanged)

    def setDialogFileMode(self, mode: qt.Qt.FileDialog):
        self._dialogFileMode = mode

    def setDialogNameFilters(self, name_filters: tuple | None):
        self._dialogNameFilters = name_filters

    def _selectFile(self):
        dialog = qt.QFileDialog(self)
        dialog.setFileMode(self._dialogFileMode)
        if self._dialogNameFilters is not None:
            dialog.setNameFilters(self._dialogNameFilters)

        if not dialog.exec():
            return

        if len(dialog.selectedFiles()) > 0:
            self.setFilePath(dialog.selectedFiles()[0])

    def getFilePath(self) -> None | str:
        file_path = self._filePath.text()
        if file_path.replace(" ", "") != "":
            return file_path
        else:
            return None

    def setFilePath(self, file_path: None | str) -> None:
        if file_path is None:
            self._filePath.clear()
        else:
            self._filePath.setText(file_path)
        self.sigFileChanged.emit()



__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "08/04/2020"


from silx.gui import qt

from darfix import dtypes
from darfix.gui.utils.message import missing_dataset_msg


class MetadataWidget(qt.QMainWindow):
    def __init__(self, parent=None):
        """
        Widget used to show the metadata in a table.
        """
        super(MetadataWidget, self).__init__(parent)
        self._dataset = None
        self.setWindowTitle("Metadata")
        self._dataset = None

        self._table = qt.QTableWidget()

        mainWidget = qt.QWidget(self)
        mainWidget.setLayout(qt.QVBoxLayout())
        mainWidget.layout().addWidget(self._table)

        self.mainWidget = mainWidget
        self.setCentralWidget(mainWidget)

    def setDataset(self, dataset: dtypes.Dataset):
        self._dataset = dataset.dataset
        self._updateView()

    def clearTable(self):
        self._table.clear()

    def _updateView(self):
        """
        Updates the view to show the correponding metadata.

        """
        if self._dataset is None:
            missing_dataset_msg()
            return

        self._table.clear()
        metadata = self._dataset.metadata_dict

        self._table.setColumnCount(len(metadata))
        self._table.setHorizontalHeaderLabels(metadata.keys())

        row_count = 0

        for column, (key, values) in enumerate(metadata.items()):
            if row_count < len(values):
                row_count = len(values)
                self._table.setRowCount(row_count)

            for row, value in enumerate(values):
                _item = qt.QTableWidgetItem()
                _item.setText(str(value))
                _item.setFlags(qt.Qt.ItemIsEnabled | qt.Qt.ItemIsSelectable)
                self._table.setItem(row, column, _item)



__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "14/05/2020"

import logging

import numpy
from sklearn.decomposition import PCA

from darfix.decomposition.ipca import IPCA

from .base import Base

_logger = logging.getLogger(__file__)


class NICA(Base):
    """Compute the non-negative independent components of the linear generative
    model x = A * s.

    Here, x is a p-dimensional observable random vector and s is the latent
    random vector of length num_components, whose components are statistically
    independent and non-negative. The matrix X is assumed to hold n samples
    of x, stacked in columns (shape(X) = (p, n)). In practice, if
    shape(X) = (p, n) this function solves X = A * S both for S and A,
    where A is the so-called mixing matrix, with shape (p, num_components),
    and S is a (num_components, n) matrix which contains n samples of the latent
    source vector, stacked in columns.

    This function implements the method presented in:
    `Blind Separation of Positive Sources by Globally Convergent Gradient Search`
    (https://core.ac.uk/download/pdf/76988305.pdf)

    :param data: array of shape (nsamples, nfeatures).
    :param num_components: Dimension of s. Number of latent random variables.
    :param chunksize: Size of every group of samples to apply PCA to. PCA will be fit with arrays
        of shape (chunksize, nfeatures), where nfeatures is the number of features per sample.
    :type chunksize: int
    :param float lr: Learning rate of gradient descent.
    :param int max_iter: Maximum number of iterations of gradient descent.
    :param float tol: Tolerance on update at each iteration.
    :type num_components: Union[uint, None]

    """

    def __init__(self, data, num_components, chunksize=None, lr=0.03, indices=None):
        Base.__init__(self, data, num_components=num_components, indices=indices)

        self.lr = lr

        _logger.info("Whitening with PCA")
        # Whitening
        if chunksize is None:
            model = PCA(num_components, whiten=True)
            if indices is not None:
                self._Z = model.fit_transform(self.data[self.indices].T).T
            else:
                self._Z = model.fit_transform(self.data[:, :].T).T
            self._V = model.components_
        else:
            ipca = IPCA(
                self.data,
                chunksize,
                num_components,
                whiten=True,
                indices=indices,
                rowvar=False,
            )
            ipca.fit_transform()
            self._Z = ipca.W.T
            self._V = ipca.H

        if self.num_components > self.Z.shape[0]:
            self.num_components = self.Z.shape[0]

    @property
    def Z(self):
        return self._Z

    @property
    def V(self):
        return self._V

    def _init_w(self):
        self.W = numpy.eye(self.num_components)

    def _update_w(self):
        """
        Apply NICA algorithm
        """
        Y = numpy.matmul(self.W, self.Z)

        f = numpy.minimum(0, Y)
        f_Y = numpy.matmul(f, Y.transpose())
        E = (f_Y - f_Y.transpose()) / Y.shape[1]
        # Gradient descent
        self.W -= self.lr * numpy.matmul(E, self.W)
        # Symmetric orthogonalization
        M = numpy.matmul(self.W, self.W.transpose())
        vals, vecs = numpy.linalg.eig(M)
        vals, vecs = vals.real, vecs.real

        W_sqrt = vecs / numpy.sqrt(vals + 1e-6)
        W_sqrt = numpy.matmul(W_sqrt, vecs.transpose())
        self.W = numpy.matmul(W_sqrt, self.W)

    def fit_transform(self, max_iter=100, error_step=None):
        _logger.info("Applying NICA algorithm")

        Base.fit_transform(
            self, max_iter=max_iter, error_step=error_step, compute_h=False
        )

        self.H = numpy.matmul(self.W, self.Z)

        WV = numpy.matmul(self.W, self.V)
        WV_ = numpy.matmul(WV, WV.transpose())
        WV_ = numpy.linalg.inv(WV_)
        self.W = numpy.matmul(WV.transpose(), WV_)



from __future__ import annotations

import logging
from typing import Optional

import h5py

_logger = logging.getLogger(__name__)


DETECTOR_KEYWORD = r"{detector}"
SCAN_KEYWORD = r"{scan}"
FIRST_SCAN_KEYWORD = r"{first_scan}"
LAST_SCAN_KEYWORD = r"{last_scan}"

EXISTING_KEYWORDS = (
    SCAN_KEYWORD,
    FIRST_SCAN_KEYWORD,
    LAST_SCAN_KEYWORD,
    DETECTOR_KEYWORD,
)


class UnsolvablePatternError(ValueError):
    """Exception raised when a pattern cannot be solved by the DataPathFinder"""

    pass


class DataPathFinder:
    """Util class to format path from a provided pattern
    At the moment it allows the following keywords:
    * {scan}: will replace the '{scan}' by an HDF5 first level group name
    * {first_scan}: will replace the '{first_scan}' by the first HDF5 group of the list
    * {last_scan}: will replace the '{last_scan}' by the first HDF5 group of the list
    * {detector}: will try to detect automatically the dataset to be used as detector
    """

    def __init__(
        self,
        file_: str | h5py.File,
        pattern: str,
        filter_entries: tuple | None = None,
        allowed_keywords: tuple = EXISTING_KEYWORDS,
    ) -> None:
        self.allowed_keywords = allowed_keywords
        self._can_be_solved = None
        self._file = file_
        assert pattern is not None, "pattern must be defined"
        self._initial_pattern = pattern
        self._solved_pattern = None
        self._filter_entries = filter_entries
        self._update_solved_pattern()

    @property
    def file(self) -> str | h5py.File:
        return self._file

    @file.setter
    def file(self, file: str | None):
        if file is not None and not isinstance(file, str):
            raise TypeError(f"file is expected to be None or a str. Get {file}")
        self._file = file
        self._update_solved_pattern()

    @property
    def pattern(self) -> str:
        return self._initial_pattern

    @pattern.setter
    def pattern(self, pattern: str):
        assert pattern is not None
        self._initial_pattern = pattern
        self._update_solved_pattern()

    @property
    def allowed_keywords(self) -> tuple:
        return self._allowed_keywords

    @allowed_keywords.setter
    def allowed_keywords(self, keywords: tuple):
        for keyword in keywords:
            if keyword not in EXISTING_KEYWORDS:
                raise ValueError(
                    f"keyword {keyword} is invalid. Valid values are {EXISTING_KEYWORDS}"
                )
        self._allowed_keywords = keywords

    @property
    def can_be_solved(self) -> bool:
        return self._can_be_solved

    def format(
        self, scan: str | None, first_scan: str | None, last_scan: str | None
    ) -> str | None:
        """Once the class is instanciated we can call the 'format' function
        to replace keywords by the scan, first_scan, last_scan values
        """
        if not self._can_be_solved:
            return None
        if (
            scan is not None
            and SCAN_KEYWORD in self._solved_pattern
            and SCAN_KEYWORD in self.allowed_keywords
        ):
            return self._solved_pattern.format(scan=scan)
        if (
            first_scan is not None
            and FIRST_SCAN_KEYWORD in self._solved_pattern
            and FIRST_SCAN_KEYWORD in self.allowed_keywords
        ):
            return self._solved_pattern.format(first_scan=first_scan)
        if (
            last_scan is not None
            and LAST_SCAN_KEYWORD in self._solved_pattern
            and LAST_SCAN_KEYWORD in self.allowed_keywords
        ):
            return self._solved_pattern.format(last_scan=last_scan)
        else:
            return self._solved_pattern

    @staticmethod
    def format_str_for_scans_keywords(
        my_str: str, scan: str | None, first_scan: str | None, last_scan: str | None
    ):
        keyword_mapping = {
            FIRST_SCAN_KEYWORD: first_scan,
            LAST_SCAN_KEYWORD: last_scan,
            SCAN_KEYWORD: scan,
        }
        format_args = dict(
            filter(
                lambda a: a[0] in my_str,
                keyword_mapping.items(),
            )
        )
        try:
            return my_str.format(
                **{
                    key.lstrip("{").rstrip("}"): value
                    for key, value in format_args.items()
                }
            )
        except ValueError as e:
            raise UnsolvablePatternError(e)

    def _update_solved_pattern(self):
        """
        create a '_solved_pattern' from the `_initial_pattern` to allow .format to be called on it.
        """
        if self.file is None:
            self._solved_pattern = None
            return
        else:
            self._can_be_solved = True
            self._solved_pattern = self._solve_pattern(pattern=self._initial_pattern)

    def _solve_pattern(self, pattern: str) -> str:
        """
        update the pattern to solve all the different keywords like '{detector}'...
        Return the solved pattern
        """
        assert pattern is not None, "pattern shoudn't be None"
        if (
            DETECTOR_KEYWORD not in pattern
            or DETECTOR_KEYWORD not in self.allowed_keywords
        ):
            solve_detector = False
        elif pattern.endswith(DETECTOR_KEYWORD):
            solve_detector = True
            pattern = pattern.replace(DETECTOR_KEYWORD, "")
        else:
            solve_detector = False
            self._can_be_solved = False
            raise UnsolvablePatternError(
                r"'{detector}' can only be placed a the end of the data path"
            )

        if isinstance(self._file, h5py.File):
            pattern = self._solve_keywords(
                self._file, my_pattern=pattern, solve_detector=solve_detector
            )
        else:
            with h5py.File(self._file, mode="r") as h5f:
                pattern = self._solve_keywords(
                    h5f, my_pattern=pattern, solve_detector=solve_detector
                )
        return pattern

    def _solve_keywords(
        self, h5f_input: h5py.Group, my_pattern: str, solve_detector: bool
    ):
        """check of the pattern can be solved and solve all keywords one by one"""
        first_scan = get_first_group(h5f_input, filter_keys=self._filter_entries)
        last_scan = get_last_group(h5f_input, filter_keys=self._filter_entries)
        if first_scan is None or last_scan is None:
            raise UnsolvablePatternError(
                f"the given file ({h5f_input.file.filename}) does not contain any group that can be considered as scan entry"
            )

        if solve_detector:
            # solve '{detector}'
            path_to_detector_data = self.format_str_for_scans_keywords(
                my_str=my_pattern,
                scan=first_scan,
                first_scan=first_scan,
                last_scan=last_scan,
            )
            detector_group = h5f_input.get(path_to_detector_data, default=None)
            if detector_group is None:
                self._can_be_solved = False
                raise UnsolvablePatternError(
                    f"Unable to find detector root group ({path_to_detector_data}) in the file ({h5f_input.file.filename})"
                )
            detector_dataset = self.find_detector_dataset(
                group=detector_group,
            )
            if detector_dataset is None:
                raise UnsolvablePatternError(
                    f"Unable to find any detector in {path_to_detector_data}"
                )
            else:
                _logger.info(f"First found detector is {detector_dataset.name}")

            # if '{scan}' keyword requested move back from 'real' path to 'pattern file'
            if SCAN_KEYWORD in my_pattern and SCAN_KEYWORD in self.allowed_keywords:
                my_pattern = self.from_found_detector_dataset_to_pattern(
                    detector_dataset=detector_dataset.name,
                    scan_path=first_scan,
                )
            else:
                my_pattern = detector_dataset.name
            return my_pattern

        test_on_first_scan = self.format_str_for_scans_keywords(
            my_str=my_pattern,
            scan=first_scan,
            first_scan=first_scan,
            last_scan=last_scan,
        )
        assert isinstance(test_on_first_scan, str)
        self._can_be_solved = test_on_first_scan in h5f_input
        return my_pattern

    @staticmethod
    def find_detector_dataset(
        group: h5py.Group, check_nexus_metadata: bool | None = None
    ) -> Optional[h5py.Dataset]:
        """
        browse all datasets / groups in the group and return the dataset the most likely to be the detector dataset.

        :param group: HDF5 group containing all elements to check.
        :param check_nexus_metadata: policy regarding checking possible metadata.
            * If True will return the first 'data' dataset contained in a group identified as an 'NXdetector' and being 3D.
            * If False will return the first 3D dataset found (can be in a sub group if named 'data')
            * If None then will look first for detector with nexus metadata else without
        """
        if not isinstance(group, h5py.Group):
            raise ValueError(
                f"group is expected to be an instance of {h5py.Group}. Get {type(group)}"
            )

        if check_nexus_metadata is None:
            return DataPathFinder.find_detector_dataset(
                group=group, check_nexus_metadata=True
            ) or DataPathFinder.find_detector_dataset(
                group=group, check_nexus_metadata=False
            )

        for name in group.keys():
            elmt = group.get(name)
            detector_dataset = DataPathFinder.get_detector(
                elmt=elmt, check_nexus_metadata=check_nexus_metadata
            )
            if detector_dataset is not None:
                return detector_dataset
        return None

    @staticmethod
    def check_is_a_3d_dataset(dataset: h5py.Dataset | h5py.Group):
        return (
            dataset is not None
            and isinstance(dataset, h5py.Dataset)
            and dataset.ndim == 3
        )

    @staticmethod
    def get_detector(
        elmt: h5py.Dataset | h5py.Group, check_nexus_metadata: bool
    ) -> Optional[h5py.Dataset]:
        if check_nexus_metadata:
            # check for nexus compliant detector
            if (
                isinstance(elmt, h5py.Group)
                and elmt.attrs.get("NX_class", None) == "NXdetector"
            ):
                data_dataset = elmt.get("data", None)
                if DataPathFinder.check_is_a_3d_dataset(dataset=data_dataset):
                    return data_dataset
            else:
                return None
        else:
            # check root level dataset
            if isinstance(elmt, h5py.Dataset):
                if DataPathFinder.check_is_a_3d_dataset(dataset=elmt):
                    return elmt
            else:
                assert isinstance(
                    elmt, h5py.Group
                ), f"elmt is expected to be a HDF5 Group. Got type({elmt})"
                # check possible 'data' dataset contained in groups
                data_dataset = elmt.get("data", None)
                if DataPathFinder.check_is_a_3d_dataset(dataset=data_dataset):
                    return data_dataset

    @staticmethod
    def from_found_detector_dataset_to_pattern(detector_dataset: str, scan_path: str):
        """
        Recreate the 'detector_dataset' pattern like '/{scan}/path/to/detectors/groups/detector' from
        the path of the detector for a specific entry ("scan_path")
        'existing' pattern like '/{scan}/path/to/detectors/groups/detector/data' or '/{scan}/path/to/detectors/groups/detector_data'
        """
        if not isinstance(detector_dataset, str):
            raise TypeError(
                f"detector_dataset should be a str. Get {type(detector_dataset)}"
            )
        if not isinstance(scan_path, str):
            raise TypeError(f"scan_path should be a str. Get {type(scan_path)}")
        # get rid of possible initial '/'
        detector_dataset = detector_dataset.lstrip("/")
        scan_path = scan_path.lstrip("/")

        if not detector_dataset.startswith(scan_path):
            raise ValueError(
                f"'detector_dataset' ({detector_dataset}) should start by 'scan_path' ({scan_path})"
            )
        # kind of a left replace based on '/' sections. As we tested the 'scan_path' starts the string it should always work
        scan_depth = len(scan_path.split("/"))
        # replace the scan_path by a '{scan}' pattern
        new_detector_dataset = "/".join(detector_dataset.split("/")[scan_depth:])
        new_detector_dataset = "/".join((SCAN_KEYWORD, new_detector_dataset))
        return new_detector_dataset


def sort_bliss_scan_entries(entries: tuple):
    """
    Sort Bliss scans (x.y) according to their scan numbers and processed entries (`entry_xxxx`) according to their entry number.
    """

    def get_entry_scan_num(entry_name):
        # concatenation resulting output entry name is "entry_0000" so let's handle this case.
        if entry_name.startswith("entry_"):
            return int(entry_name.lstrip("entry_"))
        # entries are expected to be given as x.y (bliss policy)
        return int(entry_name.lstrip("/").split(".")[0])

    return sorted(entries, key=get_entry_scan_num)


def _get_next_group(
    h5f: h5py.Group | str,
    reverse_iteration: bool,
    filter_keys: tuple | None = None,
    as_hdf5_item: bool = False,
):
    """util to retrieve the first Group not in filtered_keys of a file."""
    if as_hdf5_item and not isinstance(h5f, h5py.Group):
        raise TypeError(
            "To return the group as an hdf5 item you must provide hf5 as h5py.Group. Else returned group will be closed with the file."
        )
    if filter_keys is not None:
        # remove the left '/' that can sometime bring troubles
        filter_keys = [filter_key.lstrip("/") for filter_key in filter_keys]

    def filter_not_group(root):
        try:
            entries = sort_bliss_scan_entries(root.keys())
        except ValueError:
            _logger.error("Failed to order scans by indices. Take them 'unordered'")
            entries = root.keys()

        if reverse_iteration:
            key_iterator = list(entries)
            key_iterator.reverse()
        else:
            key_iterator = entries
        for key in key_iterator:
            if filter_keys is not None and key.lstrip("/") not in filter_keys:
                continue
            elmt = root.get(key)
            if isinstance(elmt, h5py.Group):
                if as_hdf5_item:
                    return elmt
                else:
                    return elmt.name
        return None

    if isinstance(h5f, str):
        with h5py.File(h5f, mode="r") as root:
            return filter_not_group(root)
    else:
        return filter_not_group(h5f)


def get_first_group(
    h5f: h5py.Group | str, filter_keys: tuple | None = None, as_hdf5_item: bool = False
) -> str | None:
    return _get_next_group(
        h5f=h5f,
        filter_keys=filter_keys,
        reverse_iteration=False,
        as_hdf5_item=as_hdf5_item,
    )


def get_last_group(
    h5f: h5py.Group | str, filter_keys: tuple | None = None, as_hdf5_item: bool = False
) -> str | None:
    return _get_next_group(
        h5f=h5f,
        filter_keys=filter_keys,
        reverse_iteration=True,
        as_hdf5_item=as_hdf5_item,
    )



from __future__ import annotations

import h5py
from silx.gui import qt

from darfix.core.datapathfinder import DETECTOR_KEYWORD
from darfix.core.datapathfinder import EXISTING_KEYWORDS
from darfix.core.datapathfinder import FIRST_SCAN_KEYWORD
from darfix.core.datapathfinder import LAST_SCAN_KEYWORD
from darfix.core.datapathfinder import SCAN_KEYWORD
from darfix.core.datapathfinder import get_first_group
from darfix.core.datapathfinder import get_last_group


class _TreeBuilder:
    """Build the tree to be used by the DataPathModel"""

    def __init__(
        self,
        file: str,
        qt_root_node: qt.QStandardItem,
        root_keywords: tuple,
        leaf_keywords: tuple,
    ) -> None:
        assert isinstance(
            qt_root_node, qt.QStandardItem
        ), f"hdf5_root_node should be an instance of {qt.QStandardItem}. Got {type(qt_root_node)}"
        assert isinstance(
            root_keywords, tuple
        ), f"root_keywords should be an instance of tuple. Got {type(root_keywords)}"
        assert isinstance(
            leaf_keywords, tuple
        ), f"leaf_keywords should be an instance of tuple. Got {type(leaf_keywords)}"
        self._qt_root_node = qt_root_node
        self._file = file

        assert DETECTOR_KEYWORD not in root_keywords
        self._root_keywords = root_keywords
        self._leaf_keywords = leaf_keywords
        self._display_dataset = True

    def displayDataset(self) -> bool:
        return self._display_dataset

    def setDisplayDataset(self, display: bool):
        self._display_dataset = display

    @staticmethod
    def addFirstScanKeyword(
        h5f, root_node: qt.QStandardItem
    ) -> tuple[h5py.Group, qt.QStandardItem, bool]:
        first_group = get_first_group(h5f, as_hdf5_item=True)
        qt_root = qt.QStandardItem(FIRST_SCAN_KEYWORD)
        qt_root.setData(FIRST_SCAN_KEYWORD, _DataPathCompleter.ConcatenationRole)
        root_node.appendRow(qt_root)
        return first_group, qt_root, False

    @staticmethod
    def addLastScanKeyword(
        h5f, root_node: qt.QStandardItem
    ) -> tuple[h5py.Group, qt.QStandardItem, bool]:
        last_group = get_last_group(h5f, as_hdf5_item=True)
        qt_root = qt.QStandardItem(LAST_SCAN_KEYWORD)
        qt_root.setData(LAST_SCAN_KEYWORD, _DataPathCompleter.ConcatenationRole)
        root_node.appendRow(qt_root)
        return last_group, qt_root, False

    @staticmethod
    def addScanKeyword(
        h5f, root_node: qt.QStandardItem
    ) -> tuple[h5py.Group, qt.QStandardItem, bool]:
        first_group = get_first_group(h5f, as_hdf5_item=True)
        qt_root = qt.QStandardItem(SCAN_KEYWORD)
        qt_root.setData(SCAN_KEYWORD, _DataPathCompleter.ConcatenationRole)
        root_node.appendRow(qt_root)
        return first_group, qt_root, False

    def buildTree(self):
        """build tree from root"""

        tree_root_mapping = {
            FIRST_SCAN_KEYWORD: self.addFirstScanKeyword,
            LAST_SCAN_KEYWORD: self.addLastScanKeyword,
            SCAN_KEYWORD: self.addScanKeyword,
        }

        with h5py.File(self._file, mode="r") as h5f:
            preprocessing_info: list[tuple[qt.QStandardItem, h5py.Group, bool]] = []

            for keyword in self._root_keywords:
                hdf5_root, qt_root, append_first = tree_root_mapping[keyword](
                    h5f, self._qt_root_node
                )
                preprocessing_info.append((hdf5_root, qt_root, append_first))

            if SCAN_KEYWORD not in self._root_keywords:
                # if {scan} is part of the keywords then we don't want to add the first
                # level groups (1.1...) because they are not 'valid' options in the {scan} context
                preprocessing_info.append((h5f, self._qt_root_node, True))

            for hdf5_root, qt_root, append_first in preprocessing_info:
                self.populateItem(
                    qt_parent_item=qt_root,
                    hdf5_item=hdf5_root,
                    create_first_item=append_first,
                    display_dataset=self.displayDataset(),
                )

    def populateItem(
        self,
        hdf5_item: h5py.Group | h5py.Dataset,
        qt_parent_item: qt.QStandardItem,
        create_first_item: bool = True,
        display_dataset: bool = False,
    ):
        """
        :param create_first_item: if True then create a QStandardItem for the given hdf5_item. else skip it (in the case {scan} keyword it doesn't make sense to display all the 1.1, 2.1... groups)
        :param display_dataset: if False then the h5py.Dataset will not appear. When looking for positioners for example we want to filter those
        """
        if not display_dataset and isinstance(hdf5_item, h5py.Dataset):
            return
        if create_first_item:
            text = hdf5_item.name.split("/")[-1]
            qt_children_item = qt.QStandardItem(text)
            qt_children_item.setData(text, _DataPathCompleter.ConcatenationRole)
            qt_parent_item.appendRow(qt_children_item)
        else:
            qt_children_item = qt_parent_item

        if isinstance(hdf5_item, h5py.Dataset):
            return

        # populate from leaf keywords
        for leaf_keyword in self._leaf_keywords:
            if not self.has_already_keyword(qt_parent_item, leaf_keyword):
                qt_leaf_keyword_item = qt.QStandardItem(leaf_keyword)
                qt_leaf_keyword_item.setData(
                    leaf_keyword, _DataPathCompleter.ConcatenationRole
                )
                qt_parent_item.appendRow(qt_leaf_keyword_item)

        # populate from the hdf5 file
        for key in hdf5_item.keys():
            hdf5_children_item = hdf5_item.get(key)
            self.populateItem(
                hdf5_item=hdf5_children_item,
                qt_parent_item=qt_children_item,
                display_dataset=display_dataset,
                create_first_item=True,
            )

    @staticmethod
    def has_already_keyword(qt_item, keyword):
        for i_row in range(qt_item.rowCount()):
            for i_col in range(qt_item.columnCount()):
                child = qt_item.child(i_row, i_col)
                if child.text() == keyword:
                    return True
        return False


class DataPathModel(qt.QStandardItemModel):
    """
    Model that provide HDF5 file tree combined with some reserved keywords.
    Keywords must be provided as '{my_keyword}'.

    At the moment there is two types of keywords:
    * root keywords ({scan}, {first_scan}, {last_scan}')
    * leaf keywords ('{detector}')
    """

    def __init__(
        self,
        parent: qt.QObject | None = None,
        allowed_keywords: tuple = (),
        display_dataset: bool = True,
    ) -> None:
        self._file = None
        for pattern in allowed_keywords:
            assert (
                pattern in EXISTING_KEYWORDS
            ), f"allowed keywords are {EXISTING_KEYWORDS}"
        self._allowed_keywords = allowed_keywords
        self._display_dataset = display_dataset
        super().__init__(parent)

    def getAllowedKeywords(self) -> tuple:
        return self._allowed_keywords

    def setAllowedKeywords(self, keywords: tuple) -> None:
        self._allowed_keywords = keywords
        self.clear()
        self._build()

    def data(self, index, role):
        return super().data(index, role)

    def setFile(self, hdf5_file: str):
        self._file = hdf5_file
        self.clear()
        self._build()

    def _build(self):
        if self._file is None:
            return

        builder = _TreeBuilder(
            file=self._file,
            qt_root_node=self.invisibleRootItem(),
            root_keywords=tuple(
                filter(lambda a: a != DETECTOR_KEYWORD, self.getAllowedKeywords())
            ),
            leaf_keywords=(
                (DETECTOR_KEYWORD,)
                if DETECTOR_KEYWORD in self.getAllowedKeywords()
                else ()
            ),
        )
        builder.setDisplayDataset(display=self._display_dataset)
        builder.buildTree()


class DataPathLineEditWithCompleter(qt.QLineEdit):
    def __init__(
        self,
        parent=None,
        allowed_keywords: tuple = (),
        completer_display_dataset: bool = False,
    ):
        super().__init__(parent)

        self._completer = _DataPathCompleter(
            allowed_keywords=allowed_keywords, display_dataset=completer_display_dataset
        )
        self.setCompleter(self._completer)
        self.setToolTip("Ctrl+Space will trigger auto-completion")

    def setFile(self, file):
        self._completer.setFile(file)

    def setAllowedKeywords(self, keywords: tuple):
        self._completer.setAllowedKeywords(keywords=keywords)

    def keyPressEvent(self, key_event):
        # add Ctrl + space for completion
        modifiers = key_event.modifiers()
        if qt.Qt.KeyboardModifier.ControlModifier == modifiers:
            if key_event.key() == qt.Qt.Key.Key_Space:
                self._completer.complete(self.rect())

        super().keyPressEvent(key_event)


class _DataPathCompleter(qt.QCompleter):
    ConcatenationRole = qt.Qt.UserRole + 1

    def __init__(
        self,
        parent=None,
        allowed_keywords=EXISTING_KEYWORDS,
        display_dataset: bool = True,
    ):
        super().__init__(parent)
        self._separator = "/"
        self._model = DataPathModel(
            allowed_keywords=allowed_keywords, display_dataset=display_dataset
        )
        self.setModel(self._model)

    def setFile(self, file: str):
        self._model.setFile(file)

    def setAllowedKeywords(self, keywords: tuple):
        self._model.setAllowedKeywords(keywords=keywords)

    def pathFromIndex(self, ix):
        datalist = []
        while ix.isValid():
            datalist.insert(
                0, self._model.data(ix, _DataPathCompleter.ConcatenationRole)
            )
            ix = ix.parent()

        return self._separator.join(datalist)

    def splitPath(self, path):
        return path.split(self._separator)



from silx.gui import qt


def show_error_msg(txt: str):
    """
    Show a dialog with an error message and a OK button
    """
    qt.QMessageBox.critical(
        None, "Error", txt, buttons=qt.QMessageBox.StandardButton.Ok
    )


def missing_dataset_msg():
    """
    show a dialog to the user notifying that the current widget has no dataset
    (and as a consequence cannot process)
    """
    msg = qt.QMessageBox()
    msg.setIcon(qt.QMessageBox.Critical)
    msg.setText("No dataset defined. Unable to process")
    msg.setWindowTitle("Missing dataset")
    msg.exec()


def unable_to_search_inside_hdf5_file(err_msg: str):
    """
    Show dialog to the user to notifying that we cannot search inside a HDF5 file.
    The issue is described in the 'err_msg'
    """
    msg = qt.QMessageBox()
    msg.setIcon(qt.QMessageBox.Critical)
    msg.setText(err_msg)
    msg.setWindowTitle("Fail HDF5 group / dataset search")
    msg.setStandardButtons(qt.QMessageBox.Ok)
    msg.exec()



from __future__ import annotations

import logging
import os
from typing import Type

import h5py
from silx.gui import icons
from silx.gui import qt
from silx.gui.dialog.DatasetDialog import DatasetDialog
from silx.gui.dialog.GroupDialog import GroupDialog

from darfix.core.datapathfinder import DataPathFinder
from darfix.core.datapathfinder import UnsolvablePatternError
from darfix.core.datapathfinder import get_first_group
from darfix.core.datapathfinder import get_last_group
from darfix.gui.utils.data_path_completer import DataPathLineEditWithCompleter
from darfix.gui.utils.message import unable_to_search_inside_hdf5_file

_logger = logging.getLogger(__name__)


class DataPathSelection(qt.QGroupBox):
    """
    Widget to help user select HDF5 data paths
    """

    sigPatternChanged = qt.Signal(str)
    """"
    Emit when the data path (pattern) change. Can be an empty string in the case no pattern was defined
    """

    def __init__(
        self,
        parent: qt.QWidget | None = None,
        pattern: str | None = None,
        title: str = "data path",
        allowed_keywords: tuple = (),
        completer_display_dataset: bool = False,
        data_path_type: Type[h5py.Group] | Type[h5py.Dataset] = h5py.Dataset,
    ) -> None:
        """
        :param pattern: data path pattern
        :param title: title to give to this QGroupBox
        :param allowed_keywords: keywords that can be used by the data path finder
        :param completer_display_dataset: if True then completer will not display path to dataset. Default: False.
        """
        super().__init__(parent)
        self.setLayout(qt.QGridLayout())

        if pattern is None:
            pattern = ""

        # callback to call to generate the first example
        self._dataPathFinder = DataPathFinder(
            file_=None,
            pattern=pattern,
        )
        self._defaultPattern = pattern

        if data_path_type not in (h5py.Group, h5py.Dataset):
            raise TypeError(
                f"data_path_type should be in {(h5py.Group, h5py.Dataset)}. Got {type(self._data_path_type)}"
            )
        self._data_path_type = data_path_type

        self.setTitle(title)

        # pattern
        self._patternLabel = qt.QLabel(text="pattern")
        self.layout().addWidget(self._patternLabel, 0, 0, 1, 1)
        self._patternLabel.setSizePolicy(qt.QSizePolicy.Fixed, qt.QSizePolicy.Fixed)

        self._patternQLE = DataPathLineEditWithCompleter(
            allowed_keywords=allowed_keywords,
            completer_display_dataset=completer_display_dataset,
        )
        self._patternQLE.setText(pattern)
        self.layout().addWidget(self._patternQLE, 0, 1, 1, 2)
        self._patternQLE.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Fixed)
        self._dataPathStatus = qt.QLabel()
        self.layout().addWidget(self._dataPathStatus, 0, 3, 1, 1)
        self._dataPathStatus.setSizePolicy(qt.QSizePolicy.Fixed, qt.QSizePolicy.Fixed)

        # reset
        style = qt.QApplication.style()
        self._resetPB = qt.QPushButton(
            "reset", icon=style.standardIcon(qt.QStyle.SP_DialogResetButton)
        )
        self._resetPB.setSizePolicy(qt.QSizePolicy.Fixed, qt.QSizePolicy.Fixed)
        self.layout().addWidget(self._resetPB, 0, 4, 1, 1)

        # select
        self._select = qt.QPushButton(
            "select", icon=style.standardIcon(qt.QStyle.SP_DialogOpenButton)
        )
        self._select.setSizePolicy(qt.QSizePolicy.Fixed, qt.QSizePolicy.Fixed)
        self.layout().addWidget(self._select, 0, 5, 1, 1)

        # example solved from the first scan entry
        self._exampleLabel = qt.QLabel("example")
        self.layout().addWidget(self._exampleLabel, 1, 0, 1, 1)
        self._exampleQLE = qt.QLineEdit("")
        self._exampleQLE.setReadOnly(True)
        self._exampleQLE.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Fixed)
        self.layout().addWidget(self._exampleQLE, 1, 1, 1, 2)

        # connect signal / slot
        self._patternQLE.editingFinished.connect(self._updateDataFinderPattern)
        self._patternQLE.editingFinished.connect(self._patternChanged)
        self._resetPB.released.connect(self._resetPattern)
        self._select.released.connect(self._selectPattern)

        # set up
        self._updateExample()

    def _patternChanged(self):
        self.sigPatternChanged.emit(self._patternQLE.text())

    def _updateStatusIcon(self):
        style = qt.QApplication.style()

        if self._dataPathFinder.file is None:
            icon = style.standardIcon(qt.QStyle.SP_TitleBarContextHelpButton)
            tooltip = "no file set"
        elif self._dataPathFinder.can_be_solved:
            icon = style.standardIcon(qt.QStyle.SP_DialogYesButton)
            tooltip = "data path is valid"
        else:
            tooltip = "data path is invalid"
            icon = style.standardIcon(qt.QStyle.SP_DialogNoButton)
        self._dataPathStatus.setPixmap(icon.pixmap(qt.QSize(16, 16)))
        self._dataPathStatus.setToolTip(tooltip)

    def _updateDataFinderPattern(self):
        try:
            self._dataPathFinder.pattern = self._patternQLE.text()
        except UnsolvablePatternError:
            # in the case the pattern cannot be solved.
            # Note: in this particular use case the '_dataPathStatus' will also display a specific icon.
            pass
        finally:
            self._updateExample()

    def _resetPattern(self):
        if self._defaultPattern is not None:
            self.setPattern(self._defaultPattern, store_as_default=False)
        self._updateExample()

    def _selectPattern(self) -> None:
        """
        callback when the user click on "select"
        """
        err_msg: str | None = None
        file_path = self.getInputFile()
        if not file_path:
            err_msg = "HDF5 file not defined. Please specify file path."
        elif not os.path.exists(file_path):
            err_msg = f"HDF5 file '{file_path}' doesn't exists. Please provide an existing file."
        elif not os.access(file_path, os.R_OK):
            err_msg = f"You have no read rights on '{file_path}'."

        if err_msg:
            unable_to_search_inside_hdf5_file(err_msg=err_msg)
            return

        if self._data_path_type is h5py.Group:
            dialog = GroupDialog()
        elif self._data_path_type is h5py.Dataset:
            dialog = DatasetDialog()
        else:
            raise RuntimeError(
                f"data_path_type should be in {(h5py.Group, h5py.Dataset)}. Got {type(self._data_path_type)}"
            )
        dialog.addFile(file_path)
        if dialog.exec():
            url = dialog.getSelectedDataUrl()
            self.setPattern(pattern=url.data_path(), store_as_default=False)

    def setLabelExample(self, label: str):
        self._exampleLabel.setText(label)

    def setPattern(self, pattern: str | None, store_as_default: bool = False):
        # in the case we don't want metadata for example
        # we might want the pattern to be None
        if pattern is None:
            pattern = ""
        if not isinstance(pattern, str):
            raise TypeError(
                f"pattern should be an instance of str. Got {type(pattern)}"
            )
        self._patternQLE.setText(pattern)
        try:
            self._updateDataFinderPattern()
        except Exception as e:
            _logger.error(f"Failed to update data finder. Error is {e}")
        if store_as_default:
            self._defaultPattern = pattern
        self.sigPatternChanged.emit(self.getPattern())

    def getPattern(self) -> str:
        pattern = self._patternQLE.text()
        if pattern.replace(" ", "") == "":
            # make sure if the user just enter spaces processing will be run as expected
            pattern = ""
        return pattern

    def setDefaultPattern(self, pattern: str):
        "The default pattern is only used to reset the pattern use"
        self._defaultPattern = pattern

    def getDefaultPattern(self) -> str:
        return self._defaultPattern

    def setInputFile(self, file_path: str):
        """
        :param file_path: path to the file
        :param update_path: if true then will call callback to automatically determine the expected data path
        """
        try:
            self._dataPathFinder.file = file_path
        except Exception as e:
            _logger.error(f"Failed to update file path. Error is {e}")
        finally:
            self._patternQLE.completer().setFile(file_path)
            self._updateExample()

    def getInputFile(self) -> str | None:
        return self._dataPathFinder.file

    def _updateExample(self):
        missing_finder_information = []
        if self._dataPathFinder.file is None:
            missing_finder_information.append("no input file provided")
        if self._dataPathFinder.pattern is None:
            missing_finder_information.append("no pattern to search provided")
        missing_inputs = len(missing_finder_information) > 0
        tooltip = ""
        if not missing_inputs:
            first_scan = get_first_group(self._dataPathFinder.file)
            last_scan = get_last_group(self._dataPathFinder.file)
            try:
                example = self._dataPathFinder.format(
                    scan=first_scan, first_scan=first_scan, last_scan=last_scan
                )
            except Exception as e:
                err_mess = f"Failed to compute example. Error is {e}"
                _logger.error(err_mess)
                example = None
                tooltip = err_mess
            else:
                if example is None:
                    tooltip = f"Unable to find any dataset / group that could solve the pattern ({self._dataPathFinder.pattern}) with the current file ({self._dataPathFinder.file})"
        else:
            example = None
            tooltip = f"Missing input for the data path finder: {','.join(missing_finder_information)}"

        if example is None:
            self._exampleQLE.setText("")
        else:
            self._exampleQLE.setText(example)
        self._exampleQLE.setToolTip(tooltip)
        self._updateStatusIcon()

    def getExample(self) -> str:
        return self._exampleQLE.text()

    def _getWarningIcon(self):
        return icons.getQIcon("darfix:gui/icons/warning")

    def hideExample(self):
        self._exampleLabel.hide()
        self._exampleQLE.hide()

    # expose API
    def setPlaceholderText(self, text):
        self._patternQLE.setPlaceholderText(text)

    def setAvailableKeywords(self, keywords: tuple) -> None:
        self._dataPathFinder.allowed_keywords = keywords
        self._patternQLE.setAllowedKeywords(keywords=keywords)



import os
from typing import NamedTuple

import numpy
import pytest
from silx.io.dictdump import nxtodict

from darfix.tasks.hdf5_scans_concatenation import ConcatenateHDF5Scans
from darfix.tasks.hdf5_scans_concatenation import _concatenate_dict
from darfix.tasks.hdf5_scans_concatenation import guess_output_file
from darfix.tests.utils import create_scans


class DetectorDataParameter(NamedTuple):
    raw_data_path: str
    data_path_pattern: str


@pytest.mark.parametrize("overwrite", (True, False))
@pytest.mark.parametrize(
    "detector_data_path_and_pattern",
    (
        DetectorDataParameter(
            "{scan}/measurement/my_detector", "{scan}/measurement/{detector}"
        ),
        DetectorDataParameter(
            "{scan}/instrument/my_detector/data", "{scan}/instrument/{detector}"
        ),
    ),
)
def test_scan_concatenation(tmp_path, detector_data_path_and_pattern, overwrite: bool):
    """
    creates 'n' scans and concatenate them together.
    Then check that `ConcatenateHDF5Scans` task is correctly handling it.

    :param detector_data_path: path to the detector. Can contain the full path or the '{detector}' keyword. In this case search of the detector path will be done automatically.
    """
    raw_detector_data_path, pattern = detector_data_path_and_pattern
    test_folder = tmp_path / "test_concatenation"
    test_folder.mkdir()
    input_file = test_folder / "raw_data.hdf5"
    output_file = test_folder / "concatenate_data.hdf5"

    create_scans(file_path=input_file, detector_path=raw_detector_data_path)

    task = ConcatenateHDF5Scans(
        inputs={
            "input_file": input_file,
            "output_file": output_file,
            "detector_data_path": pattern,
            "overwrite": overwrite,
        }
    )
    task.run()

    assert os.path.exists(output_file)
    output_dict = nxtodict(
        h5file=str(output_file),
    )

    def get_nx_dict_key(ddict: dict, data_path):
        for key in data_path.split("/"):
            ddict = ddict.get(key, {})
        return ddict

    output_detector_dataset = get_nx_dict_key(
        output_dict,
        data_path=raw_detector_data_path.format(scan="entry_0000"),
    )

    assert isinstance(
        output_detector_dataset, numpy.ndarray
    ), f"'output_detector_dataset' should be a numpy array. Get {type(output_detector_dataset)}"

    # make sure the detector has been properly concatenated
    raw_detector_dataset = numpy.linspace(0, 5, 100 * 100 * 4).reshape(4, 100, 100)
    numpy.testing.assert_almost_equal(
        output_detector_dataset,
        numpy.concatenate(
            (
                raw_detector_dataset,
                raw_detector_dataset,
                raw_detector_dataset,
            )
        ),
    )
    # make sure the positioners group has been properly concatenated
    output_positioners_group = get_nx_dict_key(
        output_dict,
        ConcatenateHDF5Scans.DEFAULT_POSITIONERS_DATA_PATH.format(scan="entry_0000"),
    )

    # if there is a single value along all the dataset then we save a unique value in the final dataset
    assert output_positioners_group["alpha"] == 1.0
    numpy.testing.assert_array_almost_equal(
        output_positioners_group["beta"],
        numpy.concatenate(
            [
                numpy.arange(4, dtype=numpy.float32),
                numpy.arange(4, dtype=numpy.float32),
                numpy.arange(4, dtype=numpy.float32),
            ]
        ),
    )
    numpy.testing.assert_array_almost_equal(
        output_positioners_group["gamma"],
        numpy.concatenate(
            [
                numpy.linspace(68, 70, 4, dtype=numpy.uint8),
                numpy.linspace(68, 70, 4, dtype=numpy.uint8),
                numpy.linspace(68, 70, 4, dtype=numpy.uint8),
            ]
        ),
    )
    # if there is not enough points (not one per frame) then we concatenate it anyway (but with a warning)
    numpy.testing.assert_array_almost_equal(
        output_positioners_group["delta"],
        numpy.concatenate(
            [
                numpy.arange(2, dtype=numpy.int16),
                numpy.arange(2, dtype=numpy.int16),
                numpy.arange(2, dtype=numpy.int16),
            ]
        ),
    )

    # check that according to the 'overwrite' parameter we can overwrite the file or not
    if overwrite:
        task.run()
    else:
        with pytest.raises(OSError):
            task.run()


@pytest.mark.parametrize("target_processed_data_dir", (True, False))
def test_guess_output_file(target_processed_data_dir):
    """
    Test 'guess_output_file' function. This function is used to determine a 'default' output path for a scan that we want to concatenate.
    """
    input_file = os.path.sep.join(("path", "to", "file.hdf5"))
    expected_output_file = os.path.sep.join(("path", "to", "file_darfix_concat.hdf5"))
    assert (
        guess_output_file(
            input_file=input_file, target_processed_data_dir=target_processed_data_dir
        )
        == expected_output_file
    )

    input_file = os.path.sep.join(("path", "to", "RAW_DATA", "file.nx"))
    if target_processed_data_dir:
        expected_output_file = os.path.sep.join(
            ("path", "to", "PROCESSED_DATA", "file_darfix_concat.nx")
        )
    else:
        expected_output_file = os.path.sep.join(
            ("path", "to", "RAW_DATA", "file_darfix_concat.nx")
        )
    assert (
        guess_output_file(
            input_file=input_file, target_processed_data_dir=target_processed_data_dir
        )
        == expected_output_file
    )


def test__concatenate_dict():
    """test _concatenate_dict function"""
    res = _concatenate_dict(
        {
            "a": numpy.zeros(2),
            "b": numpy.zeros(2),
        },
        {
            "a": numpy.ones(2),
            "c": numpy.ones(2),
        },
    )
    assert tuple(res.keys()) == ("a", "b", "c")
    numpy.testing.assert_array_equal(
        res["a"], numpy.concatenate([numpy.zeros(2), numpy.ones(2)])
    )
    numpy.testing.assert_array_equal(res["b"], numpy.zeros(2))
    numpy.testing.assert_array_equal(res["c"], numpy.ones(2))



import numpy

from darfix.decomposition.nmf import NMF
from darfix.tests.decomposition.utils import images
from darfix.tests.decomposition.utils import sampler


def test_fit_transform():
    rstate = numpy.random.RandomState(1000)
    resources = ["circle", "star", "pentagon", "square"]
    num_images = 100
    means = [15, 30, 45, 60]
    tol = 0.5
    sample = sampler(resources, means)

    X = numpy.array([sample(i).flatten() for i in range(num_images)])

    nmf = NMF(X, 4)
    H = rstate.random((nmf.num_components, nmf.num_features)) + 10**-4
    W = rstate.random((nmf.num_samples, nmf.num_components)) + 10**-4
    nmf.fit_transform(max_iter=500, H=H, W=W)

    stack = numpy.asarray(images(resources))
    for img in stack:
        img = img / numpy.linalg.norm(img)
        found = False
        for component in nmf.H:
            comp = component.reshape(img.shape)
            comp = comp / numpy.linalg.norm(comp)
            err = numpy.linalg.norm(comp - img)
            if err < tol:
                found = True
                break
        assert found is True



from __future__ import annotations

from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread

from darfix.dtypes import Dataset
from darfix.gui.projectionWidget import ProjectionWidget
from darfix.tasks.projection import Projection


class ProjectionWidgetOW(OWEwoksWidgetOneThread, ewokstaskclass=Projection):
    """
    Removes one dimension by projecting (summing) all images in this dimension.

    Details in https://gitlab.esrf.fr/XRD/darfix/-/issues/37
    """

    name = "projection"
    # icon = "icons/projection.png"
    want_main_area = False

    _ewoks_inputs_to_hide_from_orange = ("dimension",)

    def __init__(self):
        super().__init__()

        self._widget = ProjectionWidget(parent=self)
        self._widget.sigProjectButtonClicked.connect(
            self.execute_ewoks_task_without_propagation
        )
        self._widget.sigDimensionsChanged.connect(self._update_dimension)
        self._widget.sigOkClicked.connect(self._accept_result)
        self._get_control_layout().addWidget(self._widget)

        dataset: Dataset = self.get_default_input_value("dataset", None)
        if dataset is not None:
            self._widget.setDataset(dataset.dataset)
        # TODO: Deal with dimension default input??

    def _update_dimension(self):
        self.set_default_input("dimension", self._widget.getDimension())

    def _accept_result(self):
        self.propagate_downstream()
        self.close()

    def handleNewSignals(self) -> None:
        dataset: Dataset = self.get_task_input_value("dataset", None)
        if dataset is None:
            return
        self._widget.setDataset(dataset.dataset)
        self.open()

        # Do not call super().handleNewSignals() to prevent propagation

    def task_output_changed(self):
        send_dataset = self.get_task_output_value("dataset", None)
        if send_dataset is None:
            return
        assert isinstance(send_dataset, Dataset)
        self._widget.updatePlot(send_dataset.dataset)



import numpy

from darfix.dtypes import Dataset
from darfix.tasks.binning import Binning
from darfix.tests.utils import createDataset


def test_binning(tmp_path):
    data = numpy.array(
        [
            [
                [1, 2, 3, 4],
                [2, 2, 3, 4],
                [3, 2, 3, 4],
                [4, 2, 3, 4],
            ],
            [
                [5, 6, 7, 8],
                [5, 6, 7, 8],
                [5, 6, 7, 8],
                [5, 6, 7, 8],
            ],
            [
                [9, 10, 11, 12],
                [9, 10, 11, 12],
                [9, 10, 11, 12],
                [9, 10, 11, 12],
            ],
        ]
    )

    dataset = createDataset(data)

    task = Binning(
        inputs={
            "dataset": Dataset(dataset),
            "scale": 0.5,
        }
    )
    task.run()

    new_dataset = task.outputs.dataset.dataset

    numpy.testing.assert_array_equal(
        new_dataset.data,
        [
            [
                [1, 3],
                [2, 3],
            ],
            [
                [5, 7],
                [5, 7],
            ],
            [
                [9, 11],
                [9, 11],
            ],
        ],
    )



from __future__ import annotations

import numpy
import silx.math
from silx.utils.enum import Enum as _Enum


class Method(_Enum):
    """
    Methods available to compute the background.
    """

    MEDIAN = "median"
    MEAN = "mean"


def compute_background(
    bg_frames: numpy.ndarray,
    method: str | Method,
) -> numpy.ndarray:
    """Compute a background image from a stack of frames"""
    method = Method(method)
    if method is Method.MEAN:
        bg = numpy.mean(bg_frames, axis=0)
    elif method is Method.MEDIAN:
        bg = numpy.median(bg_frames, axis=0)
    else:
        raise NotImplementedError(f"method {method.value} not Implemented yet")

    bg = bg.astype(bg_frames.dtype)

    return bg


def background_subtraction(img: numpy.ndarray, bg: numpy.ndarray) -> None:
    """
    Compute background subtraction.

    :param array_like img: Raw image
    :param array_like bg: Background image

    :return: Image with subtracted background
    """
    # Substract and replace negative value by zero
    img[img < bg] = 0
    numpy.subtract(img, bg, out=img, where=(img > 0))


def hot_pixel_removal(image: numpy.ndarray, ksize: int = 3) -> None:
    """
    Function to remove hot pixels of the data using median filter.

    :param data: Input data.
    :param ksize: Size of the mask to apply.
    """
    median = silx.math.medfilt(image, ksize)
    if numpy.issubdtype(image.dtype, numpy.integer):
        subtracted_image = numpy.subtract(image, median, dtype=numpy.int32)
    else:
        subtracted_image = numpy.subtract(image, median)
    threshold = numpy.std(subtracted_image)
    hot_pixels = subtracted_image > threshold
    image[hot_pixels] = median[hot_pixels]


def threshold_removal(
    data: numpy.ndarray, bottom: int | None = None, top: int | None = None
) -> None:
    """
    Set bottom and top threshold to the images in the dataset.

    :param array_like data: Input data
    :param int bottom: Bottom threshold
    :param int top: Top threshold
    :returns: ndarray
    """

    if bottom is not None:
        data[data < bottom] = 0
    if top is not None:
        data[data > top] = 0


def mask_removal(data: numpy.ndarray, mask: numpy.ndarray) -> None:
    """
    Set 0 values of mask to 0.

    :param array_like data: Input data
    :param nd.array mask: Input mask.

    :returns: ndarray with the masked values
    """
    numpy.multiply(data, mask, out=data)



from __future__ import annotations

from silx.gui import qt
from silx.gui.dialog.DatasetDialog import DatasetDialog
from silx.gui.dialog.GroupDialog import GroupDialog


class _BaseLineEdit(qt.QWidget):
    """
    A line edit for paths that can filled via a selection dialog.

    The dialog creation and result must be implemented in `_getDialogResult`
    """

    dialogSelected = qt.Signal(str)
    editingFinished = qt.Signal()

    def __init__(self, parent: qt.QWidget | None = None, **kwargs) -> None:
        super().__init__(parent, **kwargs)

        self._lineEdit = qt.QLineEdit()
        browseButton = qt.QPushButton("Browse...")

        layout = qt.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._lineEdit)
        layout.addWidget(browseButton)

        browseButton.clicked.connect(self._openDialog)
        self._lineEdit.editingFinished.connect(self.editingFinished)

    def _getDialogResult(self) -> str | None:
        raise NotImplementedError()

    def _openDialog(self):
        result = self._getDialogResult()
        if not result:
            return

        self._lineEdit.setText(result)
        self.dialogSelected.emit(result)

    def setText(self, text: str):
        self._lineEdit.setText(text)

    def getText(self) -> str:
        return self._lineEdit.text()


class FileLineEdit(_BaseLineEdit):
    """A line edit for file paths that can filled via a file selection dialog"""

    dialogSelected = qt.Signal(str)
    editingFinished = qt.Signal()

    def _getDialogResult(self) -> None | str:
        dialog = qt.QFileDialog()
        result = dialog.exec()

        if not result:
            return None
        return dialog.selectedFiles()[0]


class DatasetLineEdit(_BaseLineEdit):
    """
    A line edit for an HDF5 dataset path that can filled via a dataset selection dialog.
    The file needs to be set beforehand with `setFile`
    """

    dialogSelected = qt.Signal(str)
    editingFinished = qt.Signal()

    def __init__(self, parent: qt.QWidget | None = None, **kwargs) -> None:
        super().__init__(parent, **kwargs)
        self._file = None

    def setFile(self, file: str):
        self._file = file

    def _getDialogResult(self) -> None | str:
        dialog = DatasetDialog()
        if self._file is None:
            return None

        dialog.addFile(self._file)
        result = dialog.exec()
        if not result:
            return None

        url = dialog.getSelectedDataUrl()
        return url.data_path() if url else None


class GroupLineEdit(_BaseLineEdit):
    """
    A line edit for an HDF5 group path that can filled via a group selection dialog.
    The file needs to be set beforehand with `setFile`
    """

    dialogSelected = qt.Signal(str)
    editingFinished = qt.Signal()

    def __init__(self, parent: qt.QWidget | None = None, **kwargs) -> None:
        super().__init__(parent, **kwargs)
        self._file = None

    def setFile(self, file: str):
        self._file = file

    def _getDialogResult(self) -> None | str:
        dialog = GroupDialog()
        if self._file is None:
            return None

        dialog.addFile(self._file)
        result = dialog.exec()
        if not result:
            return None

        url = dialog.getSelectedDataUrl()
        return url.data_path() if url else None



import os

import h5py
import numpy
import pytest
from ewoksorange.tests.conftest import qtapp  # noqa F401

from darfix.dtypes import Dataset
from darfix.gui.utils.qsignalspy import QSignalSpy
from orangecontrib.darfix.widgets.hdf5dataselection import HDF5DataSelectionWidgetOW


@pytest.mark.skipif(QSignalSpy is None, reason="Unable to import QSignalSpy")
def test_HDF5DataSelectionWidgetOW(tmp_path, qtapp):  # noqa F811
    raw_data_dir = tmp_path / "raw_data"
    raw_data_dir.mkdir()

    raw_data_file = os.path.join(raw_data_dir, "raw.hdf5")
    with h5py.File(raw_data_file, mode="w") as h5f:
        detector = h5f.create_group("/1.1/instrument/detector/")
        detector["image"] = numpy.arange(0, 100 * 100 * 20).reshape(20, 100, 100)
        detector.attrs["NX_class"] = "NXdetector"
        detector["type"] = "lima"
        h5f["/1.1/measurement/detector"] = h5py.SoftLink(detector.name)

    window = HDF5DataSelectionWidgetOW()
    window._rawDataSelection.setFilePath(raw_data_file)
    expected_inputs = {
        "raw_input_file": raw_data_file,
        "raw_detector_data_path": "/1.1/measurement/detector",
        "raw_metadata_path": "/1.1/instrument/positioners",
    }
    assert window.getInputs() == expected_inputs

    window._rawDataSelection.setDetectorPath("/1.1/instrument/detector/image")
    expected_inputs["raw_detector_data_path"] = "/1.1/instrument/detector/image"
    assert window.getInputs() == expected_inputs

    # this was one of the API question. What to provide if
    # the user don't want to load metadata (like for dark).
    # We went for an empty string "". Because from the task class
    # point of view the most common would be to use the default pattern.
    window._rawDataSelection.setPositionersPath("")
    expected_inputs["raw_metadata_path"] = ""
    assert window.getInputs() == expected_inputs

    window._workflowTitleLineEdit.setText("my workflow title")
    window.saveInputs()
    expected_inputs["workflow_title"] = "my workflow title"
    assert window.getInputs() == expected_inputs

    assert window.task_succeeded is None

    waiter = QSignalSpy(window.task_executor.finished)
    window.execute_ewoks_task()
    # wait for the task_executor to be finished
    waiter.wait(5000)

    dataset = window.get_task_output_value("dataset")
    assert isinstance(dataset, Dataset)



import logging
from typing import Iterable
from typing import Optional

from silx.gui import qt

from darfix.core.noiseremoval import NoiseRemovalOperation
from darfix.core.noiseremoval import Operation
from darfix.processing.imageOperations import Method

_logger = logging.getLogger(__file__)

_KERNEL_SIZES = (3, 5)


class ParametersWidget(qt.QWidget):
    def __init__(self, parent=None):
        """Widget containing the input parameters for the noise removal operations."""
        super().__init__(parent)
        self._layout = qt.QGridLayout()
        titleFont = qt.QFont()
        titleFont.setBold(True)

        # Background subtraction
        bsLabel = qt.QLabel("Background Subtraction")
        bsLabel.setFont(titleFont)
        self._layout.addWidget(bsLabel, 0, 0, 1, 2)
        self.bsMethodsCB = qt.QComboBox(self)
        for method in Method.values():
            self.bsMethodsCB.addItem(method)
        self.bsBackgroundCB = qt.QComboBox(self)
        self.computeBS = qt.QPushButton("Compute")
        methodLabel = qt.QLabel("Method:")
        bgLabel = qt.QLabel("Background:")
        methodLabel.setMargin(0)

        self._layout.addWidget(methodLabel, 1, 0, 1, 1)
        self._layout.addWidget(bgLabel, 2, 0, 1, 1)
        self._layout.addWidget(self.bsMethodsCB, 1, 1, 1, 1)
        self._layout.addWidget(self.bsBackgroundCB, 2, 1, 1, 1)
        self._layout.addWidget(self.computeBS, 4, 1, 1, 1)
        self.computeBS.hide()

        # Hot pixel removal
        hpLabel = qt.QLabel("Hot Pixel Removal")
        hpLabel.setFont(titleFont)
        self._layout.addWidget(hpLabel, 0, 2, 1, 2)
        ksizeLabel = qt.QLabel("Kernel size:")
        self._layout.addWidget(ksizeLabel, 1, 2, 1, 1)
        self.hpSizeCB = qt.QComboBox(self)
        for size in _KERNEL_SIZES:
            self.hpSizeCB.addItem(str(size))
        self.computeHP = qt.QPushButton("Compute")
        self._layout.addWidget(self.hpSizeCB, 1, 3, 1, 1)
        self._layout.addWidget(self.computeHP, 4, 3, 1, 1)
        self.computeHP.hide()

        # Threshold removal
        tpLabel = qt.QLabel("Threshold Removal")
        tpLabel.setFont(titleFont)
        self._layout.addWidget(tpLabel, 0, 4, 1, 2)
        bottomLabel = qt.QLabel("Bottom threshold:")
        topLabel = qt.QLabel("Top threshold:")
        self._layout.addWidget(bottomLabel, 1, 4, 1, 1)
        self._layout.addWidget(topLabel, 2, 4, 1, 1)
        self.bottomLE = qt.QLineEdit()
        self.bottomLE.setValidator(qt.QIntValidator())
        self.bottomLE.setSizePolicy(qt.QSizePolicy.Ignored, qt.QSizePolicy.Preferred)
        self.topLE = qt.QLineEdit()
        self.topLE.setValidator(qt.QIntValidator())
        self.topLE.setSizePolicy(qt.QSizePolicy.Ignored, qt.QSizePolicy.Preferred)
        self.computeTP = qt.QPushButton("Compute")
        self._layout.addWidget(self.bottomLE, 1, 5, 1, 1)
        self._layout.addWidget(self.topLE, 2, 5, 1, 1)
        self._layout.addWidget(self.computeTP, 4, 5, 1, 1)
        self.computeTP.hide()

        # Mask removal
        mrLabel = qt.QLabel("Mask Removal")
        mrLabel.setFont(titleFont)
        self._layout.addWidget(mrLabel, 0, 6, 1, 2)
        maskLabel = qt.QLabel("Use mask from toolbox.\n Set values off mask to 0.")
        self.computeMR = qt.QPushButton("Compute")
        self._layout.addWidget(maskLabel, 1, 6, 1, 2)
        self._layout.addWidget(self.computeMR, 4, 7, 1, 1)
        self.computeMR.hide()

        self._layout.setHorizontalSpacing(10)
        self.setLayout(self._layout)
        self.set_default_values()

    def set_default_values(
        self, history: Optional[Iterable[NoiseRemovalOperation]] = None
    ):
        self.bottomLE.setText("0")
        self.topLE.setText("100")
        if not history:
            return
        # follow history to obtain last values set by the user
        for operation in history:
            try:
                if operation["type"] is Operation.BS:
                    index = self.bsMethodsCB.findText(
                        str(operation["parameters"]["method"])
                    )
                    if index > 0:
                        self.bsMethodsCB.setCurrentIndex(index)
                elif operation["type"] is Operation.HP:
                    index = self.hpSizeCB.findText(
                        str(operation["parameters"]["kernel_size"])
                    )
                    if index > 0:
                        self.hpSizeCB.setCurrentIndex(index)
                elif operation["type"] is Operation.THRESHOLD:
                    self.bottomLE.setText(str(operation["parameters"]["bottom"]))
                    self.topLE.setText(str(operation["parameters"]["top"]))
            except KeyError as e:
                _logger.debug(e)
                _logger.debug(f"A key is missing in {operation}")



from __future__ import annotations

import os
import string

import h5py
import numpy
from ewokscore import Task
from ewokscore.missing_data import MISSING_DATA
from ewokscore.missing_data import MissingData
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict

from darfix import dtypes


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    dataset: dtypes.Dataset
    """Input dataset containing a stack of images."""
    nvalue: float | MissingData = MISSING_DATA
    """Increase or decrease the top threshold by this fixed value."""
    indices: list[int] | MissingData = MISSING_DATA
    """Indices of the images to process. If not provided, all images will be processed."""
    title: str | MissingData = MISSING_DATA
    """Title for the output file. If not provided, title is empty."""


class WeakBeam(
    Task,
    input_model=Inputs,
    output_names=["dataset"],
):
    """
    Obtain dataset with filtered weak beam and recover its Center of Mass.
    Save file with this COM for further processing.
    """

    def run(self):
        dataset = self.inputs.dataset
        indices = self.get_input_value("indices", None)
        if isinstance(dataset, dtypes.Dataset):
            dataset = dataset.dataset
        if not isinstance(dataset, dtypes.ImageDataset):
            raise TypeError("dataset is expected to be an instance")

        nvalue = self.inputs.nvalue
        wb_dataset = dataset.recover_weak_beam(nvalue, indices=indices)
        com = wb_dataset.apply_moments(indices=indices)[0][0]
        os.makedirs(dataset.dir, exist_ok=True)
        filename = os.path.join(dataset.dir, "weakbeam_{}.hdf5".format(nvalue))

        title = self.get_input_value("title", "")
        # title can be set to None, MISSING_DATA or an empty string. So safer to use the following line
        title = title or self.get_random_title()
        with h5py.File(filename, "a") as _file:
            _file[title] = com

        self.outputs.dataset = dtypes.Dataset(
            dataset=dataset,
            indices=indices,
            bg_indices=self.inputs.dataset.bg_indices,
            bg_dataset=self.inputs.dataset.bg_dataset,
        )

    @staticmethod
    def get_random_title() -> str:
        letters = string.ascii_lowercase
        return "".join(numpy.random.choice(list(letters)) for i in range(6))



from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot.StackView import StackViewMainWindow

import darfix
from darfix import dtypes
from darfix.core.state_of_operation import Operation
from darfix.gui.utils.message import missing_dataset_msg


class BinningWidget(qt.QMainWindow):
    """
    Widget to bin the data for fastest processing
    """

    sigComputed = qt.Signal()
    """Emit once the user has validated the binned dataset"""

    sigApply = qt.Signal()
    """Emit when the user ask to apply the binning"""

    sigScaleChanged = qt.Signal(float)
    """Emit when the scale changed"""

    sigAbort = qt.Signal()
    """Emit when user request to abort processing"""

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)

        self._display_tooltip_msg = True

        widget = qt.QWidget()
        layout = qt.QGridLayout()

        self._dataset = None
        # original dataset, the one treated. Keep it to be able to re-apply modifications
        self._update_dataset = None
        # dataset with applied modifications. This is the one displayed
        self.indices = None
        self.bg_indices = None
        self.bg_dataset = None

        self._scaleLE = qt.QLineEdit("1")
        validator = qt.QDoubleValidator()
        self._scaleLE.setValidator(validator)
        _buttons = qt.QDialogButtonBox(parent=self)
        self._okB = _buttons.addButton(_buttons.Ok)
        self._applyB = _buttons.addButton(_buttons.Apply)
        self._abortB = _buttons.addButton(_buttons.Abort)
        self._resetB = _buttons.addButton(_buttons.Reset)
        self._abortB.hide()

        self._applyB.released.connect(self._applyBinning)
        self._okB.released.connect(self.apply)
        self._resetB.released.connect(self.resetStack)
        self._abortB.released.connect(self.sigAbort)

        self._sv = StackViewMainWindow()
        self._sv.setKeepDataAspectRatio(True)
        self._sv.setColormap(
            Colormap(
                name=darfix.config.DEFAULT_COLORMAP_NAME,
                normalization=darfix.config.DEFAULT_COLORMAP_NORM,
            )
        )
        layout.addWidget(qt.QLabel("Scale: "), 0, 0)
        layout.addWidget(self._scaleLE, 0, 1)
        layout.addWidget(self._sv, 1, 0, 1, 2)
        layout.addWidget(_buttons, 2, 0, 1, 2)
        widget.setLayout(layout)
        self.setCentralWidget(widget)

        # connect signal / slot
        self._scaleLE.editingFinished.connect(self._scaleChanged)

    @property
    def scale(self) -> float:
        return float(self._scaleLE.text())

    @scale.setter
    def scale(self, scale: float):
        if scale != self.scale:
            self._scaleLE.setText(str(scale))
            self._scaleChanged()

    def _scaleChanged(self, *args, **kwargs):
        self.sigScaleChanged.emit(self.scale)

    def updateResultDataset(self, dataset: dtypes.ImageDataset):
        self._update_dataset = dataset
        self.setStack(self._update_dataset)

    def setDataset(self, dataset: dtypes.Dataset):
        self._dataset = dataset.dataset
        self._update_dataset = dataset.dataset
        self.indices = dataset.indices
        self.bg_indices = dataset.bg_indices
        self.bg_dataset = dataset.bg_dataset
        self.setStack()
        if self._display_tooltip_msg:
            self.display_tooltip_msg()

    def display_tooltip_msg(self):
        msg = qt.QMessageBox()
        msg.setIcon(qt.QMessageBox.Information)
        msg.setText(
            "Binning can be used to reduce the computation time of the operations in the workflow.\n"
            + "The scale is the factor to which the images will be rescaled.\n"
            + "After the correct parameters are found, you can remove the binning widget from the workflow"
            + " and execute it either with the GUI or using ewoks.\nBinning should be applied after any"
            + " ROI to have original images size and not the binned one.\n"
        )
        msg.setWindowTitle("Fit succeeded!")
        msg.setStandardButtons(qt.QMessageBox.Ok)
        msg.exec()

    def setStack(self, dataset=None):
        """
        Sets new data to the stack.
        Mantains the current frame showed in the view.

        :param Dataset dataset: if not None, data set to the stack will be from the given dataset.
        """
        if dataset is None:
            dataset = self._dataset
        nframe = self._sv.getFrameNumber()
        self._sv.setStack(dataset.as_array3d())
        self._sv.setFrameNumber(nframe)

    def _startComputation(self):
        self._applyB.setEnabled(False)
        self._okB.setEnabled(False)

    def _endComputation(self):
        self._applyB.setEnabled(True)
        self._okB.setEnabled(True)
        self._abortB.hide()
        self._abortB.setEnabled(True)

    def _applyBinning(self):
        if self._dataset is None:
            missing_dataset_msg()
            return
        self._startComputation()
        self.sigApply.emit()

    def abort(self):
        self._abortB.setEnabled(False)
        self._update_dataset.stop_operation(Operation.BINNING)

    def apply(self):
        self.sigComputed.emit()

    def resetStack(self):
        """
        Restores stack with the dataset data.
        """
        self._update_dataset = self._dataset
        self.setStack(self._dataset)

    def clearStack(self):
        """
        Clears stack.
        """
        self._sv.setStack(None)



from __future__ import annotations

import numpy
from ewokscore import Task
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict

from darfix import dtypes


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    dataset: dtypes.Dataset
    """Input dataset containing a stack of images."""
    selected_axis: int | None = None
    """Selected axis in dataset.dataset.data. The zsum is computed for each value of this axis."""


class ZSum(
    Task,
    input_model=Inputs,
    output_names=["zsum"],
):
    """Sum all images of the dataset or images at specified indices along a given dimension."""

    def run(self):
        inputs = Inputs(**self.get_input_values())
        dataset = inputs.dataset.dataset

        if inputs.selected_axis is None:
            # AS 3D array
            self.outputs.zsum = dataset.zsum()[numpy.newaxis, :, :]
        else:
            self.outputs.zsum = dataset.z_sum_along_axis(inputs.selected_axis)



from __future__ import annotations

from ewoksorange.gui.parameterform import block_signals
from silx.gui import qt
from silx.utils.enum import Enum as _Enum


class Value(_Enum):
    """"""

    PIXEL_2X = 3.25
    PIXEL_10X = 0.65


class Orientation(_Enum):
    VERTICAL = "Vertical"
    HORIZONTAL = "Horizontal"


class MagnificationWidget(qt.QWidget):
    """
    Widget to apply magnification transformation to the data axes.
    """

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        layout = qt.QVBoxLayout()
        self._magnificationFactorWidget = _MagnificationFactorWidget(parent=self)
        layout.addWidget(self._magnificationFactorWidget)

        self._topographyCheckbox = qt.QCheckBox("Topography (obpitch)")
        self._centerAxesCheckbox = qt.QCheckBox("Center axes")
        self._centerAxesCheckbox.setChecked(True)
        self._orientationCB = qt.QComboBox()
        self._orientationCB.addItems(Orientation.values())
        topographyAxis = qt.QLabel("Topography axis: ")

        layout.addWidget(self._topographyCheckbox, alignment=qt.Qt.AlignRight)
        self._topographyWidget = qt.QWidget()
        topographyLayout = qt.QHBoxLayout()
        topographyLayout.addWidget(topographyAxis)
        topographyLayout.addWidget(self._orientationCB)
        self._topographyWidget.setLayout(topographyLayout)
        self._topographyWidget.hide()
        self._topographyWidget.setMaximumHeight(40)
        layout.addWidget(self._topographyWidget, alignment=qt.Qt.AlignRight)
        layout.addWidget(self._centerAxesCheckbox)

        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        layout.addWidget(spacer)

        self._topographyCheckbox.toggled.connect(self._topographyWidget.setVisible)

        self.setLayout(layout)

    @property
    def magnification(self):
        return self._magnificationFactorWidget.getMagnification()

    @property
    def orientation(self):
        if self._topographyCheckbox.isChecked():
            return self._orientationCB.currentIndex()
        else:
            return -1

    @orientation.setter
    def orientation(self, orientation):
        if orientation != -1:
            self._topographyCheckbox.setChecked(True)
            self._orientationCB.setCurrentIndex(orientation)

    @magnification.setter
    def magnification(self, magnification: float):
        self._magnificationFactorWidget.setMagnification(magnification=magnification)


class _MagnificationFactorWidget(qt.QWidget):
    """
    Widget to define a magnification
    """

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        layout = qt.QGridLayout()

        layout.addWidget(qt.QLabel("magnification: ", self), 0, 0, 2, 1)

        self._magnificationQSB = qt.QDoubleSpinBox(parent=self)
        self._magnificationQSB.setMinimum(0)
        self._magnificationQSB.setSingleStep(0.1)
        self._magnificationQSB.setDecimals(2)
        layout.addWidget(self._magnificationQSB, 0, 1, 2, 1)

        buttons_font = self.font()
        buttons_font.setPixelSize(10)

        self._2xMagnificationRB = qt.QPushButton("2x magnification")
        layout.addWidget(
            self._2xMagnificationRB,
            0,
            2,
            1,
            1,
        )
        self._2xMagnificationRB.setCheckable(True)
        self._2xMagnificationRB.setFont(buttons_font)
        self._2xMagnificationRB.setFocusPolicy(qt.Qt.NoFocus)

        self._10xMagnificationRB = qt.QPushButton("10x magnification")
        layout.addWidget(
            self._10xMagnificationRB,
            1,
            2,
            1,
            1,
        )
        self._10xMagnificationRB.setCheckable(True)
        self._10xMagnificationRB.setFont(buttons_font)
        self._10xMagnificationRB.setFocusPolicy(qt.Qt.NoFocus)

        # connect signal / slot
        self._2xMagnificationRB.released.connect(
            lambda: self.setMagnification(Value.PIXEL_2X.value)
        )

        self._10xMagnificationRB.released.connect(
            lambda: self.setMagnification(Value.PIXEL_10X.value)
        )
        self._magnificationQSB.editingFinished.connect(self._updateCheckedButton)

        # set up
        self.setMagnification(Value.PIXEL_2X.value)
        self.setLayout(layout)

    def getMagnification(self) -> float:
        return self._magnificationQSB.value()

    def setMagnification(self, magnification: float | str):
        magnification = float(magnification)
        self._magnificationQSB.setValue(magnification)
        self._updateCheckedButton()

    def _updateCheckedButton(self):
        magnification = self.getMagnification()
        with block_signals(self._2xMagnificationRB):
            self._2xMagnificationRB.setChecked(magnification == Value.PIXEL_2X.value)

        with block_signals(self._10xMagnificationRB):
            self._10xMagnificationRB.setChecked(magnification == Value.PIXEL_10X.value)



import os

import h5py
import numpy
import pytest
from silx.io.url import DataUrl

import darfix.resources.tests
from darfix.core.dataset import ImageDataset
from darfix.core.dimension import AcquisitionDims
from darfix.core.dimension import Dimension
from darfix.dtypes import Dataset
from darfix.tasks.dimensiondefinition import get_dimensions_error
from darfix.tests.utils import createHDF5Dataset


def test_find_dimensions(dataset):
    """Tests the correct finding of the dimensions"""

    dataset.find_dimensions()
    assert dataset.dims.ndim == 3
    assert dataset.dims.get(0).name == "m"
    assert dataset.dims.get(1).name == "z"
    assert dataset.dims.get(2).name == "obpitch"


def test_find_dimension_silicon_111_reflection(tmp_path, resource_files):
    """
    Test 'find_dimension' with a bunch of motor position over a real use cases that used to bring troubles.
    """
    silicon_111_reflection_file = resource_files(darfix.resources.tests).joinpath(
        os.path.join("dimensions_definition", "silicon_111_reflection.h5")
    )

    raw_motor_values = {}
    with h5py.File(silicon_111_reflection_file, mode="r") as h5f:
        raw_motor_values["chi"] = h5f["positioners/chi"][()]
        raw_motor_values["mu"] = h5f["positioners/mu"][()]

    data_folder = tmp_path / "test_fitting"
    data_folder.mkdir()
    data_file_url = DataUrl(
        file_path=os.path.join(str(data_folder), "data.h5"),
        data_path="data",
        scheme="silx",
    )
    number_of_points = 1891
    with h5py.File(data_file_url.file_path(), mode="w") as h5f:
        h5f["data"] = numpy.random.random(number_of_points).reshape(
            number_of_points, 1, 1
        )

    dataset = Dataset(
        dataset=ImageDataset(
            detector_url=data_file_url.path(),
            metadata_url=DataUrl(
                file_path=str(silicon_111_reflection_file),
                data_path="positioners",
                scheme="silx",
            ).path(),
            _dir=None,
        )
    )
    image_dataset = dataset.dataset

    # with a tolerance of 10e-9 we won't find 1081 steps over 2 dimensions
    assert len(image_dataset.dims) == 0
    image_dataset.find_dimensions(tolerance=1e-9)
    assert len(image_dataset.dims) == 2
    assert (
        numpy.prod([val.size for val in image_dataset.dims.values()]) > number_of_points
    )

    image_dataset.find_dimensions(tolerance=1e-5)
    assert (
        numpy.prod([val.size for val in image_dataset.dims.values()])
        == number_of_points
    )
    for dim in image_dataset.dims.values():
        numpy.testing.assert_almost_equal(
            dim.start, min(raw_motor_values[dim.name]), decimal=3
        )
        numpy.testing.assert_almost_equal(
            dim.stop, max(raw_motor_values[dim.name]), decimal=3
        )


def test_find_dimension_NiTi_1PD_002_g411_420MPa_mosalayers_2x(
    tmp_path, resource_files
):
    """
    Test 'find_dimension' with a bunch of motor position over a real use cases that used to bring troubles.
    """
    dataset_file = resource_files(darfix.resources.tests).joinpath(
        os.path.join(
            "dimensions_definition", "NiTi_1PD_002_g411_420MPa_mosalayers_2x.h5"
        )
    )

    raw_motor_values = {}
    with h5py.File(dataset_file, mode="r") as h5f:
        raw_motor_values["chi"] = h5f["positioners/chi"][()]
        raw_motor_values["diffry"] = h5f["positioners/diffry"][()]
        raw_motor_values["difftz"] = h5f["positioners/difftz"][()]

    data_folder = tmp_path / "test_fitting"
    data_folder.mkdir()
    data_file_url = DataUrl(
        file_path=os.path.join(str(data_folder), "data.h5"),
        data_path="data",
        scheme="silx",
    )
    number_of_points = 31500
    with h5py.File(data_file_url.file_path(), mode="w") as h5f:
        h5f["data"] = numpy.random.random(number_of_points).reshape(
            number_of_points, 1, 1
        )

    dataset = Dataset(
        dataset=ImageDataset(
            detector_url=data_file_url.path(),
            metadata_url=DataUrl(
                file_path=str(dataset_file),
                data_path="positioners",
                scheme="silx",
            ).path(),
            _dir=None,
        )
    )
    image_dataset = dataset.dataset

    def check_dimensions_bounds(dims: dict):
        """Make sure find_dimension is correctly fitting motor bounds"""
        for dim in dims.values():
            numpy.testing.assert_almost_equal(
                dim.start, raw_motor_values[dim.name][0], decimal=3
            )
            numpy.testing.assert_almost_equal(
                dim.stop, raw_motor_values[dim.name][-1], decimal=3
            )

    # with a tolerance of 10e-9 we won't find 1081 steps over 2 dimensions
    assert len(image_dataset.dims) == 0
    image_dataset.find_dimensions(tolerance=1e-5)
    assert len(image_dataset.dims) == 3
    check_dimensions_bounds(dims=image_dataset.dims)
    assert (
        numpy.prod([val.size for val in image_dataset.dims.values()]) > number_of_points
    )

    image_dataset.find_dimensions(tolerance=1e-4)
    assert len(image_dataset.dims) == 3
    check_dimensions_bounds(dims=image_dataset.dims)
    assert (
        numpy.prod([val.size for val in image_dataset.dims.values()])
        == number_of_points
    )


def test_are_dimensions_ok():
    data = numpy.ones(1).repeat(20).reshape(20, 1, 1)
    metadata_dict = {
        "dim1": numpy.arange(20),
        "dim2": numpy.arange(20),
        "dim3": numpy.arange(20),
    }
    dataset = createHDF5Dataset(data, metadata_dict)
    dims = AcquisitionDims()

    # len == 0
    assert "None dimension are defined." == get_dimensions_error(dataset, dims)

    dims = [
        Dimension("dim1", 5, 0, 19),
        Dimension("dim2", 2, 0, 19),
        Dimension("bad_name", 2, 0, 19),
    ]

    # bad name
    assert "bad_name" in get_dimensions_error(dataset, dims)

    dims = [
        Dimension("dim1", 5, 0, 19),
        Dimension("dim2", 2, 0, 19),
        Dimension("dim3", 2, 0, 21),
    ]

    # bad stop
    assert "21" in get_dimensions_error(dataset, dims)

    dims = [
        Dimension("dim1", 5, 0, 19),
        Dimension("dim2", 2, 0, 19),
        Dimension("dim3", 2, -2, 19),
    ]

    # bad start
    assert "-2" in get_dimensions_error(dataset, dims)

    dims = [
        Dimension("dim1", 5, 0, 19),
        Dimension("dim2", 2, 0, 19),
        Dimension("dim3", 1, 0, 19),
    ]

    # bad size
    assert "20" in get_dimensions_error(dataset, dims)

    dims = [
        Dimension("dim1", 5, 0, 19),
        Dimension("dim2", 2, 0, 19),
        Dimension("dim3", 2, 0, 19),
    ]

    # good
    assert get_dimensions_error(dataset, dims) is None


@pytest.mark.parametrize(
    "metadata_values, expected_start, expected_stop, expected_size, expected_step",
    [
        (numpy.array([10, 9, 8, 7, 6, 5, 4, 3, 2, 1]), 10, 1, 10, -1),
        (numpy.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]), 1, 10, 10, 1),
        (numpy.array([1, 5, 9]), 1, 9, 3, 4),
        (numpy.array([1, 1, 2, 2.01, 3, 3, 4, 3.98, 5, 5]), 1, 5, 5, 1),
        (numpy.array([5, 5, 4, 3.98, 3, 3, 2, 2.01, 1, 1]), 5, 1, 5, -1),
        (numpy.array([2, 2]), 2, 2, 1, 0),
    ],
)
def test_dim_linspace(
    metadata_values: numpy.ndarray,
    expected_start: float,
    expected_stop: float,
    expected_size: int,
    expected_step: float,
):
    dim = Dimension("dim")
    dim.guess_parameters(metadata_values, tolerance=0.01)

    # Check start, stop, and length

    numpy.testing.assert_allclose(
        dim.compute_unique_values(),
        metadata_values[:: len(metadata_values) // expected_size],
        atol=0.01,
    )
    assert numpy.isclose(expected_start, dim.start)
    assert numpy.isclose(expected_stop, dim.stop)
    assert expected_size == dim.size
    assert expected_step == dim.step



from __future__ import annotations

import copy
import logging

from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread
from silx.gui import qt

from darfix.core.utils import OperationAborted
from darfix.gui.utils.standardButtonBox import StandardButtonBox

_logger = logging.getLogger(__name__)


class OperationWidgetBase(OWEwoksWidgetOneThread, openclass=True):

    def __init__(
        self,
        buttonTypes: qt.QDialogButtonBox.StandardButton = qt.QDialogButtonBox.StandardButton.NoButton,
    ):
        super().__init__()

        self.__inputDataset = None

        # buttons
        self.__buttons = StandardButtonBox(parent=self, additionalButtons=buttonTypes)

        self.__buttons.okButton.clicked.connect(self.onOkClicked)
        self.__buttons.resetButton.clicked.connect(self.onResetClicked)
        self.__buttons.abortButton.clicked.connect(self.onAbortClicked)
        self.__buttons.okButton.setEnabled(False)

        self.task_executor.finished.connect(self.__onFinishTask)

        self.__widget = self.createMainWidget(self.get_default_input_values())
        self.mainArea.layout().addWidget(self.__widget)
        self.mainArea.layout().addWidget(self.__buttons)

    @property
    def buttons(self) -> StandardButtonBox:
        return self.__buttons

    @property
    def mainWidget(self) -> qt.QWidget:
        return self.__widget

    def saveMainWidget(self) -> None:
        """
        inherited class need to override this method.

        Save parameters as default inputs with `set_default_input` here.
        """
        raise NotImplementedError("This is an abstract method.")

    def createMainWidget(self, default_inputs: dict) -> qt.QWidget:
        """
        inherited class need to override this method.

        Instantiate and setup main widget here.

        :param default_inputs: Dictionary containing the default inputs that are set.
        :return : An instance of the created main widget
        """
        raise NotImplementedError("This is an abstract method.")

    def onOkClicked(self) -> None:
        if len(self.get_task_outputs()) == 0:
            # if no output at all this is an unexpected behaviour
            raise RuntimeError("Cannot go to next step because outputs are empty.")
        self.close()
        self.propagate_downstream()

    def onResetClicked(self):
        self.reset()

    def onAbortClicked(self):
        self.cancel_running_task()

    def __onFinishTask(self):
        self.mainWidget.setEnabled(True)
        self.buttons.setIsComputing(False)

    def handleNewSignals(self) -> None:
        # Do not call super().handleNewSignals() to prevent propagation
        self.__inputDataset = self.get_task_input_value("dataset")
        self.set_dynamic_input("dataset", copy.deepcopy(self.__inputDataset))

    def reset(self):
        self.__buttons.okButton.setEnabled(False)
        self.set_dynamic_input("dataset", self.__inputDataset)
        # Re trigger handleNewSignals to reset dataset input
        self.handleNewSignals()

    def task_output_changed(self) -> None:
        if self.task_succeeded:
            self.__buttons.okButton.setEnabled(True)

    def isAborted(self) -> bool:
        return isinstance(self.task_exception, OperationAborted)

    def closeEvent(self, evt) -> None:
        super().closeEvent(evt)
        self.saveMainWidget()

    def execute_ewoks_task_without_propagation(self):
        self.buttons.setIsComputing(True)
        self.mainWidget.setDisabled(True)
        super().execute_ewoks_task_without_propagation()



from __future__ import annotations

import logging

import numpy
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot import Plot2D
from silx.gui.plot import StackView
from silx.gui.plot.items.image import ImageStack

import darfix

from ...core.dataset import ImageDataset
from ...core.dimension import Dimension
from ...core.rocking_curves import fit_1d_rocking_curve
from ...core.rocking_curves import fit_2d_rocking_curve
from ...core.utils import TooManyDimensionsForRockingCurvesError
from ...dtypes import Dataset
from ..utils.axis_type_combobox import AxisTypeComboBox
from .utils import compute_contours

_logger = logging.getLogger(__file__)

FIT_IMAGE_LEGEND = "fit_image"
DATA_MARKER_LEGEND = "data_marker"
FIT_MARKER_LEGEND = "fit_marker"


class RockingCurvesPlot(qt.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._sv = StackView(parent=self, position=True)
        self._sv.setKeepDataAspectRatio(True)
        self._sv.setColormap(Colormap(name=darfix.config.DEFAULT_COLORMAP_NAME))
        self._sv.setLabels(("Frame number", "X pixels", "Y pixels"))

        self._fitPlot = Plot2D(parent=self)
        self._fitPlot.setDefaultColormap(
            Colormap(
                name=darfix.config.DEFAULT_COLORMAP_NAME,
                normalization=darfix.config.DEFAULT_COLORMAP_NORM,
            )
        )
        self._fitPlot.setGraphTitle("Rocking curves")

        self._axisTypeComboBox = AxisTypeComboBox()
        self._axisTypeComboBox.currentTextChanged.connect(self._replotCurves)
        self._lastFitParamsLabel = qt.QLabel("")

        layout = qt.QHBoxLayout(self)
        layout.addWidget(self._sv)

        fitPlotLayout = qt.QVBoxLayout()
        fitPlotLayout.addWidget(self._fitPlot)
        fitPlotLayout.addWidget(self._lastFitParamsLabel)
        comboBoxLayout = qt.QHBoxLayout()
        comboBoxLayout.addWidget(qt.QLabel("Axis values"))
        comboBoxLayout.addWidget(self._axisTypeComboBox)
        fitPlotLayout.addLayout(comboBoxLayout)

        layout.addLayout(fitPlotLayout)

        self._dataset: ImageDataset | None = None
        self._indices = None

    def setDataset(self, dataset: Dataset):
        self._dataset = dataset.dataset
        self._indices = dataset.indices

        self._sv.getPlotWidget().sigPlotSignal.connect(self._onClickOnStack)
        self._sv.sigFrameChanged.connect(self._replotCurves)

        self._sv.setGraphTitle(self._dataset.title)

    def updateStack(self):
        nframe = self._sv.getFrameNumber()

        if self._dataset is None:
            stack = None
        else:
            stack = self._dataset.as_array3d(self._indices)

        self._sv.setStack(stack)
        self._sv.setFrameNumber(nframe)

    def _getStackDisplayedImage(self) -> numpy.ndarray | None:
        stack = self._sv.getActiveImage()
        if not isinstance(stack, ImageStack):
            return None

        stackData = stack.getStackData(copy=False)
        if stackData is None:
            return None

        return stackData[stack.getStackPosition()]

    def _onClickOnStack(self, info):
        if info["event"] != "mouseClicked":
            return

        # In case the user has clicked on a pixel in the stack
        px = info["x"]
        py = info["y"]
        image = self._getStackDisplayedImage()
        if image is not None:
            # Show vertical and horizontal lines for clicked pixel
            xlim, ylim = image.shape
            self._sv.getPlotWidget().addCurve(
                (px, px), (0, xlim), legend="x", color="r"
            )
            self._sv.getPlotWidget().addCurve(
                (0, ylim), (py, py), legend="y", color="r"
            )
        self._plotRockingCurves(px, py)

    def _replotCurves(self, i=None):
        xc = self._sv.getPlotWidget().getCurve("x")
        if xc is None:
            return
        px = xc.getXData()[0]
        py = self._sv.getPlotWidget().getCurve("y").getYData()[0]
        self._plotRockingCurves(px, py)

    def _addFitMarker(self, x: float, y: float):
        self._fitPlot.addMarker(x, y, symbol="o", legend=FIT_MARKER_LEGEND, color="r")

    def _addDataMarker(self, x: float, y: float):
        fit_image = self._fitPlot.getImage(FIT_IMAGE_LEGEND)
        if fit_image is None:
            self._fitPlot.addMarker(
                x, y, symbol="o", legend=DATA_MARKER_LEGEND, color="b"
            )
            return

        x0, y0 = fit_image.getOrigin()
        xscale, yscale = fit_image.getScale()
        # Shift by xscale / 2 and yscale / 2 to land in the middle of the pixel
        self._fitPlot.addMarker(
            x0 + xscale / 2 + x * xscale,
            y0 + yscale / 2 + y * yscale,
            symbol="o",
            legend=DATA_MARKER_LEGEND,
            color="b",
        )

    def _addFitContours(self, contours: list[numpy.ndarray]):
        fit_image = self._fitPlot.getImage(FIT_IMAGE_LEGEND)
        if fit_image is not None:
            x0, y0 = fit_image.getOrigin()
            xscale, yscale = fit_image.getScale()
        else:
            x0, y0 = 0, 0
            xscale, yscale = 1, 1
        for i, contour in enumerate(contours):
            # Shift by xscale / 2 and yscale / 2 to land in the middle of the pixel

            self._fitPlot.addCurve(
                x=x0 + xscale / 2 + contour[:, 1] * xscale,
                y=y0 + yscale / 2 + contour[:, 0] * yscale,
                linestyle="-",
                linewidth=2.0,
                legend=f"Contour {i}",
                resetzoom=False,
                color="w",
            )

    def _computeFitPlotOrigin(
        self, xdim: Dimension, ydim: Dimension
    ) -> tuple[float, float]:
        axis_type = self._axisTypeComboBox.getCurrentAxisType()
        if axis_type == "dims":
            return (xdim.start, ydim.start)

        if axis_type == "center":
            # Center by shifting by half the scaled size
            xscale, yscale = self._computeFitPlotScale(xdim, ydim)
            return (-xscale * (xdim.size / 2), -yscale * (ydim.size / 2))

        return (0.0, 0.0)

    def _computeFitPlotScale(
        self, xdim: Dimension, ydim: Dimension
    ) -> tuple[float, float]:
        return (xdim.step, ydim.step)

    def _plotRockingCurves(self, px: float, py: float):
        """
        Plot rocking curves of data and fitted data at pixel (px, py).

        :param Data data: stack of images to plot
        :param px: x pixel
        :param py: y pixel
        """
        # Get rocking curves from data
        self._fitPlot.clear()
        if self._dataset is None:
            return

        try:
            data = self._dataset.as_array3d(self._indices)
            if isinstance(data, numpy.ndarray):
                rocking_curve = data[:, int(py), int(px)]
            else:
                _logger.warning("Data is not an array!")
                return
        except IndexError:
            _logger.warning("Index out of bounds")
            return

        if self._dataset.dims.ndim == 2:
            self._plot2DRockingCurve(self._dataset, rocking_curve)
        elif self._dataset.dims.ndim == 1:
            self._plot1DRockingCurve(self._dataset, rocking_curve)
        else:
            raise TooManyDimensionsForRockingCurvesError()

    def _plot2DRockingCurve(self, dataset: ImageDataset, rocking_curve: numpy.ndarray):
        y_values = numpy.zeros(dataset.nframes)
        y_values[self._indices] = rocking_curve
        dim1 = dataset.dims.get(1)
        dim0 = dataset.dims.get(0)
        assert dim1 is not None
        assert dim0 is not None

        origin = self._computeFitPlotOrigin(dim1, dim0)
        scale = self._computeFitPlotScale(dim1, dim0)
        self._fitPlot.remove(kind="curve")
        try:
            y_gauss, pars = fit_2d_rocking_curve(
                (y_values, None),
                x_values=[
                    dataset.get_metadata_values(key=dim0.name),
                    dataset.get_metadata_values(key=dim1.name),
                ],
                shape=(dim0.size, dim1.size),
            )
            if numpy.array_equal(y_gauss, y_values):
                raise RuntimeError()
            y_gauss = numpy.reshape(y_gauss, (dim1.size, dim0.size)).T

        except (TypeError, RuntimeError):
            _logger.warning("Cannot fit")
        else:
            self._lastFitParamsLabel.setText(
                "PEAK_X:{:.3f} PEAK_Y:{:.3f} FWHM_X:{:.3f} FWHM_Y:{:.3f} AMP:{:.3f} CORR:{:.3f} BG:{:.3f}".format(
                    *pars
                )
            )
            image = numpy.reshape(y_values, (dim1.size, dim0.size)).T
            self._fitPlot.addImage(
                image,
                xlabel=dim1.name,
                ylabel=dim0.name,
                origin=origin,
                scale=scale,
                legend=FIT_IMAGE_LEGEND,
            )
            contours = compute_contours(image)
            self._addFitContours(contours)

        frameNumber = self._sv.getFrameNumber()
        self._addDataMarker(x=(frameNumber // dim0.size), y=(frameNumber % dim0.size))

    def _plot1DRockingCurve(self, dataset: ImageDataset, rocking_curve: numpy.ndarray):
        dim = dataset.dims.get(0)
        assert dim is not None
        motor_values = dataset.get_metadata_values(key=dim.name, indices=self._indices)
        if self._axisTypeComboBox.getCurrentAxisType() == "dims":
            self._fitPlot.setGraphXLabel(dim.name)
            x = motor_values
        elif self._axisTypeComboBox.getCurrentAxisType() == "center":
            self._fitPlot.setGraphXLabel("Indices")
            x = numpy.arange(len(motor_values)) - int(dim.size / 2)
        else:
            self._fitPlot.setGraphXLabel("Indices")
            x = numpy.arange(len(motor_values))

        self._fitPlot.clear()
        self._fitPlot.addCurve(x, rocking_curve, legend="data", color="b")
        self._fitPlot.setGraphYLabel("Intensity")
        i = self._sv.getFrameNumber()
        try:
            y_gauss, pars = fit_1d_rocking_curve(
                (numpy.array(rocking_curve), None), x_values=x, num_points=1000
            )
            self._lastFitParamsLabel.setText(
                "AMP:{:.3f} PEAK:{:.3f} FWHM:{:.3f} BG:{:.3f}".format(*pars)
            )
        except TypeError:
            y_gauss = rocking_curve
            _logger.warning("Cannot fit")
        else:
            x_gauss = numpy.linspace(x[0], x[-1], len(y_gauss))
            self._fitPlot.addCurve(x_gauss, y_gauss, legend="fit", color="r")
            i_gauss = i * int((len(y_gauss) - 1) / (len(x) - 1))
            self._addFitMarker(x_gauss[i_gauss], y_gauss[i_gauss])

        self._addDataMarker(x[i], rocking_curve[i])



from __future__ import annotations

from ewokscore import Task
from ewokscore.missing_data import MISSING_DATA
from ewokscore.missing_data import MissingData
from ewokscore.missing_data import is_missing_data
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict

from darfix.dtypes import Dataset


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    dataset: Dataset
    """ Input dataset containing a stack of images """
    num_components: int | MissingData = MISSING_DATA
    """Number of principal components to compute."""


class PCA(
    Task,
    input_model=Inputs,
    output_names=["vals", "dataset"],
):
    """Compute Principal Component Analysis on a Darfix dataset.

    More about PCA : https://en.wikipedia.org/wiki/Principal_component_analysis"""

    def run(self):
        dataset = self.inputs.dataset
        if not isinstance(dataset, Dataset):
            raise TypeError(
                f"dataset is expected to be an instance of {Dataset}. Got {type(dataset)} instead"
            )
        num_components = self.get_input_value("num_components")

        pca_kwargs = {"return_vals": True, "indices": dataset.indices}
        if not is_missing_data(num_components):
            pca_kwargs["num_components"] = num_components

        vals = dataset.dataset.pca(**pca_kwargs)

        self.outputs.vals = vals
        self.outputs.dataset = Dataset(
            dataset=dataset.dataset,
            indices=dataset.indices,
            bg_dataset=dataset.bg_dataset,
            bg_indices=dataset.bg_indices,
        )



from __future__ import annotations

from typing import NamedTuple
from typing import Sequence

import numpy


class FixedDimension(NamedTuple):
    """
    A data class to describe a fixed dimension. Used as a parameter for ImageStack::filter_indices().

    This is used when we want to filter the image stack by one of the dimension.

    For instance, in my dataset dimensions, i have two motors `motor_slow` and `motor_fast`

    motor_fast
    - axis = 0
    - size : 3
    - linspace : 1, 2, 3

    motor_slow
    - axis = 1
    - size : 4
    - linspace : 0.5, 1, 1.5, 2

    Let say we want images indice of my dataset only when `motor_slow` value = 1.5

    `motor_slow` axis is 1 and value 1.5 is at index 2.

    So, I can use dataset.filter_indices(FixedDimension(axis = 1, index = 2))
    """

    axis: int
    index: int

    def get_reversed_axis(self, ndim: int) -> int:
        """
        In darfix Dimension class, `axis` =  Ndim - real_data_array_axis - 1.
        This is due to legacy.
        """
        assert ndim - self.axis - 1 >= 0
        return ndim - self.axis - 1

    @staticmethod
    def from_iterable(dimension: Sequence[int | Sequence[int]]) -> FixedDimension:
        # this is not really ideal but as for now dimension type is a mess in the code i prefere do all the dirty code here
        assert len(dimension) == 2
        dim0, dim1 = dimension
        axis = dim0 if isinstance(dim0, int) else dim0[0]
        index = dim1 if isinstance(dim1, int) else dim1[0]
        return FixedDimension(axis, index)


class ImageStack:
    def __init__(self, data: numpy.ndarray):
        self._data = data

    @property
    def nframes(self) -> int:
        """
        Return number of frames
        """
        return int(numpy.prod(self.scan_shape))

    @property
    def scan_shape(self) -> tuple:
        return self.data.shape[:-2]

    @property
    def frame_shape(self) -> tuple:
        return self.data.shape[-2:]

    @property
    def data(self) -> numpy.ndarray:
        return self._data

    def as_array3d(self, indices=None) -> numpy.ndarray:
        if indices is None:
            indices = Ellipsis
        return self._data.reshape((self.nframes,) + self.frame_shape)[indices]

    def as_array2d(self, indices=None) -> numpy.ndarray:
        if indices is None:
            indices = Ellipsis
        return self._data.reshape((self.nframes, self.data.size // self.nframes))[
            indices
        ]

    def filter_indices(
        self,
        indices: numpy.ndarray | None = None,
        fixed_dimension: FixedDimension | Sequence | None = None,
    ) -> numpy.ndarray:
        """
        :return selected_indices: a list of indices filtered by input `indices` and optionally filtered with one fixed dimension `fixed_dimension`. If inputs are None return the full range of indices.
        """
        if fixed_dimension is not None and not isinstance(
            fixed_dimension, FixedDimension
        ):
            fixed_dimension = FixedDimension.from_iterable(fixed_dimension)

        all_indices = numpy.arange(self.nframes)
        selected_indices = indices if indices is not None else all_indices

        if fixed_dimension is None or len(self._data.shape) <= 3:
            return selected_indices

        # Take indices with one fixed dimension
        filtered_by_dim_indices = all_indices.reshape(self.scan_shape)
        filtered_by_dim_indices = filtered_by_dim_indices.take(
            indices=fixed_dimension.index,
            axis=fixed_dimension.get_reversed_axis(self._data.ndim - 2),
        )
        filtered_by_dim_indices = filtered_by_dim_indices.flatten()

        selected_indices = numpy.intersect1d(
            selected_indices, filtered_by_dim_indices, assume_unique=True
        )

        return selected_indices

    def get_filtered_data(
        self,
        indices: numpy.ndarray | None = None,
        fixed_dimension: FixedDimension | None = None,
    ):
        return self.as_array3d()[
            self.filter_indices(indices=indices, fixed_dimension=fixed_dimension)
        ]



from __future__ import annotations

import numpy
from silx.gui.plot import Plot2D
from silx.utils.enum import Enum as _Enum

from ...core.grainplot import MomentType
from ...core.grainplot import MultiDimMomentType
from ...core.transformation import Transformation


class MapType(_Enum):
    """
    Different maps to show
    """

    COM = MomentType.COM.value
    FWHM = MomentType.FWHM.value
    SKEWNESS = MomentType.SKEWNESS.value
    KURTOSIS = MomentType.KURTOSIS.value
    MOSAICITY = MultiDimMomentType.MOSAICITY.value


def add_image_with_transformation(
    plot: Plot2D, image: numpy.ndarray, transformation: Transformation | None = None
):
    if transformation is None:
        plot.addImage(image, xlabel="pixels", ylabel="pixels")
        return

    if transformation.rotate:
        image = numpy.rot90(image, 3)
    plot.addImage(
        image,
        origin=transformation.origin,
        scale=transformation.scale,
        xlabel=transformation.label,
        ylabel=transformation.label,
    )



import os

import h5py
import numpy
import pytest
from silx.io.dictdump import dicttonx

from darfix.tasks.hdf5_scans_concatenation import DataPathFinder


def test_find_detector_dataset(tmp_path):
    """
    test the 'find_detector_dataset' function

    Make sure that:
     * NX_class 'NXdetector' is recognized and get the first priority other any 'untagged' 3d dataset by default.
     * Any 3D dataset is consider to be a detector in the case the NX_class is ignored.
    During processing we will first look for dataset with a NX_class defined as NXdetector
    Then only look for 3D dataset
    """
    test_folder = tmp_path / "test_find_detector_dataset"
    test_folder.mkdir()

    file_path = os.path.join(test_folder, "dataset.hdf5")
    dicttonx(
        {
            "group": {
                "data": "test",
            },
            "my_2d_detector": {
                "data": numpy.linspace(0, 100, 100).reshape(10, 10),
            },
            "my_3d_detector": {
                "data": numpy.linspace(0, 100, 1000).reshape(10, 10, 10),
            },
            "my_2d_nx_detector": {
                "@NX_class": "NXdetector",
                "data": numpy.linspace(0, 100, 100).reshape(10, 10),
            },
            "my_3d_nx_detector": {
                "@NX_class": "NXdetector",
                "data": numpy.linspace(0, 100, 1000).reshape(10, 10, 10),
            },
        },
        h5file=file_path,
        h5path="entry/positioners",
    )

    with h5py.File(file_path, mode="r") as h5f:
        assert (
            DataPathFinder.find_detector_dataset(
                h5f["entry/positioners"], check_nexus_metadata=None
            )
            == h5f["entry/positioners/my_3d_nx_detector/data"]
        )
    with h5py.File(file_path, mode="r") as h5f:
        assert (
            DataPathFinder.find_detector_dataset(
                h5f["entry/positioners"], check_nexus_metadata=False
            )
            == h5f["entry/positioners/my_3d_detector/data"]
        )

    dicttonx(
        {
            "group": {
                "data": "test",
            },
            "my_2d_detector": numpy.linspace(0, 100, 100).reshape(10, 10),
            "my_3d_detector": numpy.linspace(0, 100, 1000).reshape(10, 10, 10),
        },
        h5file=file_path,
        h5path="entry/positioners",
    )

    with h5py.File(file_path, mode="r") as h5f:
        assert (
            DataPathFinder.find_detector_dataset(
                h5f["entry/positioners"], check_nexus_metadata=False
            )
            == h5f["entry/positioners/my_3d_detector"]
        )


def test_from_found_detector_dataset_to_pattern():
    """
    test 'from_found_detector_dataset_to_pattern' static function.
    This function is used during '{detector}' keyword resolution
    """
    assert (
        DataPathFinder.from_found_detector_dataset_to_pattern(
            detector_dataset="1.1/positioners/detector/data",
            scan_path="1.1",
        )
        == r"{scan}/positioners/detector/data"
    )

    assert (
        DataPathFinder.from_found_detector_dataset_to_pattern(
            detector_dataset="1.1/positioners/detector",
            scan_path="1.1",
        )
        == r"{scan}/positioners/detector"
    )

    assert (
        DataPathFinder.from_found_detector_dataset_to_pattern(
            detector_dataset="new/path/positioners/detector",
            scan_path="new/path",
        )
        == r"{scan}/positioners/detector"
    )

    with pytest.raises(ValueError):
        DataPathFinder.from_found_detector_dataset_to_pattern(
            detector_dataset="new/path/positioners/detector",
            scan_path="tot",
        )



from __future__ import annotations

from ewokscore.missing_data import is_missing_data
from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread

from darfix.dtypes import Dataset
from darfix.gui.PCAWidget import PCAPlot
from darfix.tasks.pca import PCA


class PCAWidgetOW(OWEwoksWidgetOneThread, ewokstaskclass=PCA):
    name = "PCA"
    description = "A widget to perform principal component analysis"
    icon = "icons/pca.png"
    want_main_area = True
    want_control_area = False

    _ewoks_inputs_to_hide_from_orange = "num_components"

    def __init__(self):
        super().__init__()

        self._plot = PCAPlot(parent=self)
        self.mainArea.layout().addWidget(self._plot)

    def task_output_changed(self) -> None:
        vals = self.get_task_output_value("vals")
        dataset: Dataset | None = self.get_task_output_value("dataset", None)

        if is_missing_data(vals):
            self._plot.clear()
        else:
            self._plot.setData(vals, dataset.dataset.title if dataset else None)
        return super().task_output_changed()

    def handleNewSignals(self) -> None:
        super().handleNewSignals()
        self.open()
